import pytest
import numpy as np
import numpy.testing as npt
from lenstronomy.LensModel.Profiles.nie import NIE
from lenstronomy.LensModel.Profiles.chameleon import (
    Chameleon,
    DoubleChameleon,
    DoubleChameleonPointMass,
    TripleChameleon,
)
from lenstronomy.LightModel.Profiles.chameleon import (
    DoubleChameleon as DoubleChameleonLight,
)
from lenstronomy.LightModel.Profiles.chameleon import (
    TripleChameleon as TripleChameleonLight,
)
import lenstronomy.Util.param_util as param_util


class TestChameleon(object):
    """Class to test the Moffat profile."""

    def setup_method(self):
        self.chameleon = Chameleon()
        self.nie = NIE()

    def test_theta_E_convert(self):
        w_c, w_t = 2, 1
        theta_E_convert, w_c, w_t, s_scale_1, s_scale_2 = self.chameleon.param_convert(
            alpha_1=1, w_c=w_c, w_t=w_t, e1=0, e2=0
        )
        assert w_c == 1
        assert w_t == 2
        assert theta_E_convert == 0

    def test_function(self):
        """

        :return:
        """
        x = np.linspace(0.1, 10, 10)
        w_c, w_t = 0.5, 1.0
        phi_G, q = 0.3, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_light = {"alpha_1": 1.0, "w_c": 0.5, "w_t": 1.0, "e1": e1, "e2": e2}
        theta_E_convert, w_c, w_t, s_scale_1, s_scale_2 = self.chameleon.param_convert(
            alpha_1=1, w_c=0.5, w_t=1.0, e1=e1, e2=e2
        )
        kwargs_1 = {
            "theta_E": theta_E_convert,
            "s_scale": s_scale_1,
            "e1": e1,
            "e2": e2,
        }
        kwargs_2 = {
            "theta_E": theta_E_convert,
            "s_scale": s_scale_2,
            "e1": e1,
            "e2": e2,
        }
        f_ = self.chameleon.function(x=x, y=1.0, **kwargs_light)
        f_1 = self.nie.function(x=x, y=1.0, **kwargs_1)
        f_2 = self.nie.function(x=x, y=1.0, **kwargs_2)
        npt.assert_almost_equal(f_, (f_1 - f_2), decimal=5)

    def test_derivatives(self):
        """

        :return:
        """
        x = np.linspace(0.1, 10, 10)
        w_c, w_t = 0.5, 1.0
        phi_G, q = 0.3, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_light = {"alpha_1": 1.0, "w_c": 0.5, "w_t": 1.0, "e1": e1, "e2": e2}
        theta_E_convert, w_c, w_t, s_scale_1, s_scale_2 = self.chameleon.param_convert(
            alpha_1=1, w_c=0.5, w_t=1.0, e1=e1, e2=e2
        )
        kwargs_1 = {
            "theta_E": theta_E_convert,
            "s_scale": s_scale_1,
            "e1": e1,
            "e2": e2,
        }
        kwargs_2 = {
            "theta_E": theta_E_convert,
            "s_scale": s_scale_2,
            "e1": e1,
            "e2": e2,
        }
        f_x, f_y = self.chameleon.derivatives(x=x, y=1.0, **kwargs_light)
        f_x_1, f_y_1 = self.nie.derivatives(x=x, y=1.0, **kwargs_1)
        f_x_2, f_y_2 = self.nie.derivatives(x=x, y=1.0, **kwargs_2)
        npt.assert_almost_equal(f_x, (f_x_1 - f_x_2), decimal=5)
        npt.assert_almost_equal(f_y, (f_y_1 - f_y_2), decimal=5)
        f_x, f_y = self.chameleon.derivatives(x=1, y=0.0, **kwargs_light)
        npt.assert_almost_equal(f_x, 1, decimal=1)

    def test_hessian(self):
        """

        :return:
        """
        x = np.linspace(0.1, 10, 10)
        w_c, w_t = 0.5, 1.0
        phi_G, q = 0.3, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_light = {"alpha_1": 1.0, "w_c": 0.5, "w_t": 1.0, "e1": e1, "e2": e2}
        theta_E_convert, w_c, w_t, s_scale_1, s_scale_2 = self.chameleon.param_convert(
            alpha_1=1, w_c=0.5, w_t=1.0, e1=e1, e2=e2
        )
        kwargs_1 = {
            "theta_E": theta_E_convert,
            "s_scale": s_scale_1,
            "e1": e1,
            "e2": e2,
        }
        kwargs_2 = {
            "theta_E": theta_E_convert,
            "s_scale": s_scale_2,
            "e1": e1,
            "e2": e2,
        }
        f_xx, f_xy, f_yx, f_yy = self.chameleon.hessian(x=x, y=1.0, **kwargs_light)
        f_xx_1, f_xy_1, f_yx_1, f_yy_1 = self.nie.hessian(x=x, y=1.0, **kwargs_1)
        f_xx_2, f_xy_2, f_yx_2, f_yy_2 = self.nie.hessian(x=x, y=1.0, **kwargs_2)
        npt.assert_almost_equal(f_xx, (f_xx_1 - f_xx_2), decimal=5)
        npt.assert_almost_equal(f_yy, (f_yy_1 - f_yy_2), decimal=5)
        npt.assert_almost_equal(f_xy, (f_xy_1 - f_xy_2), decimal=5)
        npt.assert_almost_equal(f_yx, (f_yx_1 - f_yx_2), decimal=5)

    def test_static(self):
        x, y = 1.0, 1.0
        phi_G, q = 0.3, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_light = {"alpha_1": 1.0, "w_c": 0.5, "w_t": 1.0, "e1": e1, "e2": e2}
        f_ = self.chameleon.function(x, y, **kwargs_light)
        self.chameleon.set_static(**kwargs_light)
        f_static = self.chameleon.function(x, y, **kwargs_light)
        npt.assert_almost_equal(f_, f_static, decimal=8)
        self.chameleon.set_dynamic()
        kwargs_light = {"alpha_1": 2.0, "w_c": 0.5, "w_t": 1.0, "e1": e1, "e2": e2}
        f_dyn = self.chameleon.function(x, y, **kwargs_light)
        assert f_dyn != f_static


class TestDoubleChameleon(object):
    """Class to test the Moffat profile."""

    def setup_method(self):
        pass

    def test_param_name(self):
        chameleon = DoubleChameleon()
        names = chameleon.param_names
        assert names[0] == "alpha_1"

    def test_function(self):
        """

        :return:
        """
        doublechameleon = DoubleChameleon()
        chameleon = Chameleon()

        x = np.linspace(0.1, 10, 10)
        phi_G, q = 0.3, 0.8
        theta_E = 1.0
        ratio = 2.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_light = {
            "alpha_1": 1.0,
            "ratio": 2,
            "w_c1": 0.5,
            "w_t1": 1.0,
            "e11": e1,
            "e21": e2,
            "w_c2": 0.1,
            "w_t2": 0.5,
            "e12": e1,
            "e22": e2,
        }

        kwargs_1 = {
            "alpha_1": theta_E / (1 + 1.0 / ratio),
            "w_c": 0.5,
            "w_t": 1.0,
            "e1": e1,
            "e2": e2,
        }
        kwargs_2 = {
            "alpha_1": theta_E / (1 + ratio),
            "w_c": 0.1,
            "w_t": 0.5,
            "e1": e1,
            "e2": e2,
        }
        flux = doublechameleon.function(x=x, y=1.0, **kwargs_light)
        flux1 = chameleon.function(x=x, y=1.0, **kwargs_1)
        flux2 = chameleon.function(x=x, y=1.0, **kwargs_2)
        npt.assert_almost_equal(flux, flux1 + flux2, decimal=8)

    def test_derivatives(self):
        """

        :return:
        """
        doublechameleon = DoubleChameleon()
        chameleon = Chameleon()

        x = np.linspace(0.1, 10, 10)
        phi_G, q = 0.3, 0.8
        theta_E = 1.0
        ratio = 2.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_light = {
            "alpha_1": 1.0,
            "ratio": 2,
            "w_c1": 0.5,
            "w_t1": 1.0,
            "e11": e1,
            "e21": e2,
            "w_c2": 0.1,
            "w_t2": 0.5,
            "e12": e1,
            "e22": e2,
        }

        kwargs_1 = {
            "alpha_1": theta_E / (1 + 1.0 / ratio),
            "w_c": 0.5,
            "w_t": 1.0,
            "e1": e1,
            "e2": e2,
        }
        kwargs_2 = {
            "alpha_1": theta_E / (1 + ratio),
            "w_c": 0.1,
            "w_t": 0.5,
            "e1": e1,
            "e2": e2,
        }
        f_x, f_y = doublechameleon.derivatives(x=x, y=1.0, **kwargs_light)
        f_x1, f_y1 = chameleon.derivatives(x=x, y=1.0, **kwargs_1)
        f_x2, f_y2 = chameleon.derivatives(x=x, y=1.0, **kwargs_2)
        npt.assert_almost_equal(f_x, f_x1 + f_x2, decimal=8)
        npt.assert_almost_equal(f_y, f_y1 + f_y2, decimal=8)

    def test_hessian(self):
        """

        :return:
        """
        doublechameleon = DoubleChameleon()
        chameleon = Chameleon()

        x = np.linspace(0.1, 10, 10)
        phi_G, q = 0.3, 0.8
        theta_E = 1.0
        ratio = 2.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens = {
            "alpha_1": theta_E,
            "ratio": ratio,
            "w_c1": 0.5,
            "w_t1": 1.0,
            "e11": e1,
            "e21": e2,
            "w_c2": 0.1,
            "w_t2": 0.5,
            "e12": e1,
            "e22": e2,
        }
        kwargs_light = {
            "amp": theta_E,
            "ratio": ratio,
            "w_c1": 0.5,
            "w_t1": 1.0,
            "e11": e1,
            "e21": e2,
            "w_c2": 0.1,
            "w_t2": 0.5,
            "e12": e1,
            "e22": e2,
        }

        kwargs_1 = {
            "alpha_1": theta_E / (1 + 1.0 / ratio),
            "w_c": 0.5,
            "w_t": 1.0,
            "e1": e1,
            "e2": e2,
        }
        kwargs_2 = {
            "alpha_1": theta_E / (1 + ratio),
            "w_c": 0.1,
            "w_t": 0.5,
            "e1": e1,
            "e2": e2,
        }
        f_xx, f_xy, f_yx, f_yy = doublechameleon.hessian(x=x, y=1.0, **kwargs_lens)
        f_xx1, f_xy1, f_yx1, f_yy1 = chameleon.hessian(x=x, y=1.0, **kwargs_1)
        f_xx2, f_xy2, f_yx2, f_yy2 = chameleon.hessian(x=x, y=1.0, **kwargs_2)
        npt.assert_almost_equal(f_xx, f_xx1 + f_xx2, decimal=8)
        npt.assert_almost_equal(f_yy, f_yy1 + f_yy2, decimal=8)
        npt.assert_almost_equal(f_xy, f_xy1 + f_xy2, decimal=8)
        npt.assert_almost_equal(f_yx, f_yx1 + f_yx2, decimal=8)
        light = DoubleChameleonLight()
        f_xx, f_xy, f_yx, f_yy = doublechameleon.hessian(
            x=np.linspace(0, 1, 10), y=np.zeros(10), **kwargs_lens
        )
        kappa = 1.0 / 2 * (f_xx + f_yy)
        kappa_norm = kappa / np.mean(kappa)
        flux = light.function(x=np.linspace(0, 1, 10), y=np.zeros(10), **kwargs_light)
        flux_norm = flux / np.mean(flux)
        npt.assert_almost_equal(kappa_norm, flux_norm, decimal=5)

    def test_static(self):
        doublechameleon = DoubleChameleon()
        x, y = 1.0, 1.0
        phi_G, q = 0.3, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_light = {
            "alpha_1": 1,
            "ratio": 0.5,
            "w_c1": 0.5,
            "w_t1": 1.0,
            "e11": e1,
            "e21": e2,
            "w_c2": 0.1,
            "w_t2": 0.5,
            "e12": e1,
            "e22": e2,
        }
        f_ = doublechameleon.function(x, y, **kwargs_light)
        doublechameleon.set_static(**kwargs_light)
        f_static = doublechameleon.function(x, y, **kwargs_light)
        npt.assert_almost_equal(f_, f_static, decimal=8)
        doublechameleon.set_dynamic()
        kwargs_light = {
            "alpha_1": 2,
            "ratio": 0.5,
            "w_c1": 0.5,
            "w_t1": 1.0,
            "e11": e1,
            "e21": e2,
            "w_c2": 0.1,
            "w_t2": 0.5,
            "e12": e1,
            "e22": e2,
        }
        f_dyn = doublechameleon.function(x, y, **kwargs_light)
        assert f_dyn != f_static


class TestDoubleChameleonPointMass(object):
    """Class to test the Moffat profile."""

    def setup_method(self):
        pass

    def test_param_name(self):
        chameleon = DoubleChameleonPointMass()
        names = chameleon.param_names
        assert names[0] == "alpha_1"

    def test_function(self):
        """

        :return:
        """
        doublechameleon = DoubleChameleonPointMass()

        phi_G, q = 0.3, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_light = {
            "alpha_1": 1.0,
            "ratio_pointmass": 3,
            "ratio_chameleon": 2,
            "w_c1": 0.5,
            "w_t1": 1.0,
            "e11": e1,
            "e21": e2,
            "w_c2": 0.1,
            "w_t2": 0.5,
            "e12": e1,
            "e22": e2,
        }
        flux = doublechameleon.function(x=1, y=1.0, **kwargs_light)
        npt.assert_almost_equal(flux, 1.2602247653486218, decimal=4)

    def test_derivatives(self):
        """

        :return:
        """
        doublechameleon = DoubleChameleonPointMass()

        phi_G, q = 0.3, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_light = {
            "alpha_1": 1.0,
            "ratio_pointmass": 3,
            "ratio_chameleon": 2,
            "w_c1": 0.5,
            "w_t1": 1.0,
            "e11": e1,
            "e21": e2,
            "w_c2": 0.1,
            "w_t2": 0.5,
            "e12": e1,
            "e22": e2,
        }
        f_x, f_y = doublechameleon.derivatives(x=1, y=1.0, **kwargs_light)
        npt.assert_almost_equal(f_x, 0.43419725313692664, decimal=4)
        npt.assert_almost_equal(f_y, 0.4521464786719726, decimal=4)

    def test_hessian(self):
        """

        :return:
        """
        doublechameleon = DoubleChameleonPointMass()

        phi_G, q = 0.3, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_light = {
            "alpha_1": 1.0,
            "ratio_pointmass": 3,
            "ratio_chameleon": 2,
            "w_c1": 0.5,
            "w_t1": 1.0,
            "e11": e1,
            "e21": e2,
            "w_c2": 0.1,
            "w_t2": 0.5,
            "e12": e1,
            "e22": e2,
        }
        f_xx, f_xy, f_yx, f_yy = doublechameleon.hessian(x=1, y=1.0, **kwargs_light)
        npt.assert_almost_equal(f_xx, 0.06255816336369684, decimal=4)
        npt.assert_almost_equal(f_xy, -0.3986532840628945, decimal=4)
        npt.assert_almost_equal(f_yx, -0.3986532840628945, decimal=4)
        npt.assert_almost_equal(f_yy, 0.04715726782095693, decimal=4)


class TestTripleChameleon(object):
    """Class to test the Moffat profile."""

    def setup_method(self):
        pass

    def test_param_name(self):
        chameleon = TripleChameleon()
        names = chameleon.param_names
        assert names[0] == "alpha_1"

    def test_function(self):
        """

        :return:
        """
        triplechameleon = TripleChameleon()
        chameleon = Chameleon()

        x = np.linspace(0.1, 10, 10)
        phi_G, q = 0.3, 0.8
        ratio12 = 2.0
        ratio13 = 3
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_light = {
            "alpha_1": 1.0,
            "ratio12": ratio12,
            "ratio13": ratio13,
            "w_c1": 0.5,
            "w_t1": 1.0,
            "e11": e1,
            "e21": e2,
            "w_c2": 0.1,
            "w_t2": 0.5,
            "e12": e1,
            "e22": e2,
            "w_c3": 0.1,
            "w_t3": 0.5,
            "e13": e1,
            "e23": e2,
        }

        amp1 = 1.0 / (1.0 + 1.0 / ratio12 + 1.0 / ratio13)
        amp2 = amp1 / ratio12
        amp3 = amp1 / ratio13
        kwargs_1 = {"alpha_1": amp1, "w_c": 0.5, "w_t": 1.0, "e1": e1, "e2": e2}
        kwargs_2 = {"alpha_1": amp2, "w_c": 0.1, "w_t": 0.5, "e1": e1, "e2": e2}
        kwargs_3 = {"alpha_1": amp3, "w_c": 0.1, "w_t": 0.5, "e1": e1, "e2": e2}
        flux = triplechameleon.function(x=x, y=1.0, **kwargs_light)
        flux1 = chameleon.function(x=x, y=1.0, **kwargs_1)
        flux2 = chameleon.function(x=x, y=1.0, **kwargs_2)
        flux3 = chameleon.function(x=x, y=1.0, **kwargs_3)
        npt.assert_almost_equal(flux, flux1 + flux2 + flux3, decimal=8)

    def test_derivatives(self):
        """

        :return:
        """
        triplechameleon = TripleChameleon()
        chameleon = Chameleon()

        x = np.linspace(0.1, 10, 10)
        phi_G, q = 0.3, 0.8
        ratio12 = 2.0
        ratio13 = 3
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_light = {
            "alpha_1": 1.0,
            "ratio12": ratio12,
            "ratio13": ratio13,
            "w_c1": 0.5,
            "w_t1": 1.0,
            "e11": e1,
            "e21": e2,
            "w_c2": 0.1,
            "w_t2": 0.5,
            "e12": e1,
            "e22": e2,
            "w_c3": 0.1,
            "w_t3": 0.5,
            "e13": e1,
            "e23": e2,
        }

        amp1 = 1.0 / (1.0 + 1.0 / ratio12 + 1.0 / ratio13)
        amp2 = amp1 / ratio12
        amp3 = amp1 / ratio13
        kwargs_1 = {"alpha_1": amp1, "w_c": 0.5, "w_t": 1.0, "e1": e1, "e2": e2}
        kwargs_2 = {"alpha_1": amp2, "w_c": 0.1, "w_t": 0.5, "e1": e1, "e2": e2}
        kwargs_3 = {"alpha_1": amp3, "w_c": 0.1, "w_t": 0.5, "e1": e1, "e2": e2}
        f_x, f_y = triplechameleon.derivatives(x=x, y=1.0, **kwargs_light)
        f_x1, f_y1 = chameleon.derivatives(x=x, y=1.0, **kwargs_1)
        f_x2, f_y2 = chameleon.derivatives(x=x, y=1.0, **kwargs_2)
        f_x3, f_y3 = chameleon.derivatives(x=x, y=1.0, **kwargs_3)
        npt.assert_almost_equal(f_x, f_x1 + f_x2 + f_x3, decimal=8)
        npt.assert_almost_equal(f_y, f_y1 + f_y2 + f_y3, decimal=8)

    def test_hessian(self):
        """

        :return:
        """
        triplechameleon = TripleChameleon()
        chameleon = Chameleon()

        x = np.linspace(0.1, 10, 10)
        phi_G, q = 0.3, 0.8
        ratio12 = 2.0
        ratio13 = 3
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens = {
            "alpha_1": 1.0,
            "ratio12": ratio12,
            "ratio13": ratio13,
            "w_c1": 0.5,
            "w_t1": 1.0,
            "e11": e1,
            "e21": e2,
            "w_c2": 0.1,
            "w_t2": 0.5,
            "e12": e1,
            "e22": e2,
            "w_c3": 0.1,
            "w_t3": 0.5,
            "e13": e1,
            "e23": e2,
        }

        kwargs_light = {
            "amp": 1.0,
            "ratio12": ratio12,
            "ratio13": ratio13,
            "w_c1": 0.5,
            "w_t1": 1.0,
            "e11": e1,
            "e21": e2,
            "w_c2": 0.1,
            "w_t2": 0.5,
            "e12": e1,
            "e22": e2,
            "w_c3": 0.1,
            "w_t3": 0.5,
            "e13": e1,
            "e23": e2,
        }

        amp1 = 1.0 / (1.0 + 1.0 / ratio12 + 1.0 / ratio13)
        amp2 = amp1 / ratio12
        amp3 = amp1 / ratio13
        kwargs_1 = {"alpha_1": amp1, "w_c": 0.5, "w_t": 1.0, "e1": e1, "e2": e2}
        kwargs_2 = {"alpha_1": amp2, "w_c": 0.1, "w_t": 0.5, "e1": e1, "e2": e2}
        kwargs_3 = {"alpha_1": amp3, "w_c": 0.1, "w_t": 0.5, "e1": e1, "e2": e2}

        f_xx, f_xy, f_yx, f_yy = triplechameleon.hessian(x=x, y=1.0, **kwargs_lens)
        f_xx1, f_xy1, f_yx1, f_yy1 = chameleon.hessian(x=x, y=1.0, **kwargs_1)
        f_xx2, f_xy2, f_yx2, f_yy2 = chameleon.hessian(x=x, y=1.0, **kwargs_2)
        f_xx3, f_xy3, f_yx3, f_yy3 = chameleon.hessian(x=x, y=1.0, **kwargs_3)
        npt.assert_almost_equal(f_xx, f_xx1 + f_xx2 + f_xx3, decimal=8)
        npt.assert_almost_equal(f_yy, f_yy1 + f_yy2 + f_yy3, decimal=8)
        npt.assert_almost_equal(f_xy, f_xy1 + f_xy2 + f_xy3, decimal=8)
        npt.assert_almost_equal(f_yx, f_yx1 + f_yx2 + f_yx3, decimal=8)
        light = TripleChameleonLight()
        f_xx, f_xy, f_yx, f_yy = triplechameleon.hessian(
            x=np.linspace(0, 1, 10), y=np.zeros(10), **kwargs_lens
        )
        kappa = 1.0 / 2 * (f_xx + f_yy)
        kappa_norm = kappa / np.mean(kappa)
        flux = light.function(x=np.linspace(0, 1, 10), y=np.zeros(10), **kwargs_light)
        flux_norm = flux / np.mean(flux)
        npt.assert_almost_equal(kappa_norm, flux_norm, decimal=5)

    def test_static(self):
        triplechameleon = TripleChameleon()
        x, y = 1.0, 1.0
        phi_G, q = 0.3, 0.8
        ratio12 = 2.0
        ratio13 = 3
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_lens = {
            "alpha_1": 1.0,
            "ratio12": ratio12,
            "ratio13": ratio13,
            "w_c1": 0.5,
            "w_t1": 1.0,
            "e11": e1,
            "e21": e2,
            "w_c2": 0.1,
            "w_t2": 0.5,
            "e12": e1,
            "e22": e2,
            "w_c3": 0.1,
            "w_t3": 0.5,
            "e13": e1,
            "e23": e2,
        }
        f_ = triplechameleon.function(x, y, **kwargs_lens)
        triplechameleon.set_static(**kwargs_lens)
        f_static = triplechameleon.function(x, y, **kwargs_lens)
        npt.assert_almost_equal(f_, f_static, decimal=8)
        triplechameleon.set_dynamic()
        kwargs_lens = {
            "alpha_1": 2.0,
            "ratio12": ratio12,
            "ratio13": ratio13,
            "w_c1": 0.5,
            "w_t1": 1.0,
            "e11": e1,
            "e21": e2,
            "w_c2": 0.1,
            "w_t2": 0.5,
            "e12": e1,
            "e22": e2,
            "w_c3": 0.1,
            "w_t3": 0.5,
            "e13": e1,
            "e23": e2,
        }
        f_dyn = triplechameleon.function(x, y, **kwargs_lens)
        assert f_dyn != f_static


if __name__ == "__main__":
    pytest.main()



__author__ = "lenstronomy developers"
__email__ = "lenstronomy-dev@googlegroups.com"
__version__ = "1.11.5"
__credits__ = "ETH Zurich, UCLA, Stanford, Stony Brook"

from .Util.package_util import short, laconic



import numpy as np
import scipy.special as special
from lenstronomy.GalKin import velocity_util
from scipy.interpolate import interp1d

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
class Anisotropy(object):
    """
    class that handles the kinematic anisotropy
    sources: Mamon & Lokas 2005
    https://arxiv.org/pdf/astro-ph/0405491.pdf

    Agnello et al. 2014
    https://arxiv.org/pdf/1401.4462.pdf
    """

    def __init__(self, anisotropy_type):
        """

        :param anisotropy_type: string, anisotropy model type
        """
        self._type = anisotropy_type
        if self._type == "const":
            self._model = Const()
        elif self._type == "radial":
            self._model = Radial()
        elif self._type == "isotropic":
            self._model = Isotropic()
        elif self._type == "OM":
            self._model = OsipkovMerritt()
        elif self._type == "GOM":
            self._model = GeneralizedOM()
        elif self._type == "Colin":
            self._model = Colin()
        else:
            raise ValueError("anisotropy type %s not supported!" % self._type)

    def beta_r(self, r, **kwargs):
        """Returns the anisotropy parameter at a given radius.

        :param r: 3d radius
        :param kwargs: parameters of the specified anisotropy model
        :return: beta(r)
        """
        return self._model.beta_r(r, **kwargs)

    def K(self, r, R, **kwargs):
        """Equation A16 im Mamon & Lokas for Osipkov&Merrit anisotropy.

        :param r: 3d radius
        :param R: projected 2d radius
        :param kwargs: parameters of the specified anisotropy model
        :return: K(r, R)
        """
        return self._model.K(r, R, **kwargs)

    def anisotropy_solution(self, r, **kwargs):
        """The solution to d ln(f)/ d ln(r) = 2 beta(r)

        :param r: 3d radius
        :param kwargs: parameters of the specified anisotropy model
        :return: f(r)
        """
        return self._model.anisotropy_solution(r, **kwargs)

    def delete_anisotropy_cache(self):
        """Deletes cached interpolations for a fixed anisotropy model.

        :return: None
        """
        if hasattr(self._model, "delete_cache"):
            self._model.delete_cache()


@export
class Const(object):
    """Constant anisotropy model class See Mamon & Lokas 2005 for details."""

    def __init__(self):
        pass

    @staticmethod
    def K(r, R, beta):
        """Equation A16 im Mamon & Lokas for constant anisotropy.

        :param r: 3d radius
        :param R: projected 2d radius
        :param beta: anisotropy, float >-0.5
        :return: K(r, R, beta)
        """
        u = r / R
        k = np.sqrt(1 - 1.0 / u**2) / (1.0 - 2 * beta) + np.sqrt(
            np.pi
        ) / 2 * special.gamma(beta - 1.0 / 2) / special.gamma(beta) * (
            3.0 / 2 - beta
        ) * u ** (
            2 * beta - 1.0
        ) * (
            1 - special.betainc(beta + 1.0 / 2, 1.0 / 2, 1.0 / u**2)
        )
        return k

    @staticmethod
    def beta_r(r, beta):
        """Anisotropy as a function of radius.

        :param r: 3d radius
        :param beta: anisotropy
        :return: beta
        """
        return beta

    def anisotropy_solution(self, r, **kwargs):
        """The solution to d ln(f)/ d ln(r) = 2 beta(r)

        :param r: 3d radius
        :param kwargs: parameters of the specified anisotropy model
        :return: f(r)
        """
        raise ValueError("routine not supported yet for constant anisotropy model!")


@export
class Isotropic(object):
    """Class for isotropic (beta=0) stellar orbits See Mamon & Lokas 2005 for
    details."""

    def __init__(self):
        pass

    @staticmethod
    def K(r, R):
        """Equation A16 im Mamon & Lokas for constant anisotropy.

        :param r: 3d radius
        :param R: projected 2d radius
        :return: K(r, R)
        """
        u = r / R
        k = np.sqrt(1 - 1.0 / u**2)
        return k

    @staticmethod
    def beta_r(r):
        """Anisotropy as a function of radius.

        :param r: 3d radius
        :return: beta
        """
        return 0.0

    @staticmethod
    def anisotropy_solution(r, **kwargs):
        """The solution to d ln(f)/ d ln(r) = 2 beta(r) See e.g. A3 in Mamon & Lokas.

        :param r: 3d radius
        :param kwargs: parameters of the specified anisotropy model
        :return: f(r)
        """
        return 1


@export
class Radial(object):
    """Class for radial (beta=1) stellar orbits See Mamon & Lokas 2005 for details."""

    def __init__(self):
        pass

    @staticmethod
    def K(r, R):
        """Equation A16 im Mamon & Lokas for constant anisotropy.

        :param r: 3d radius
        :param R: projected 2d radius
        :return: K(r, R)
        """
        u = r / R
        k = (
            np.pi / 4 * u
            - 1.0 / 2 * np.sqrt(1 - 1.0 / u**2)
            - u / 2.0 * np.arcsin(1.0 / u)
        )
        return k

    @staticmethod
    def beta_r(r):
        """Anisotropy as a function of radius.

        :param r: 3d radius
        :return: beta
        """
        return 1.0

    @staticmethod
    def anisotropy_solution(r):
        """The solution to d ln(f)/ d ln(r) = 2 beta(r) See e.g. A4 in Mamon & Lokas.

        :param r: 3d radius
        :return: f(r)
        """
        return r**2


@export
class OsipkovMerritt(object):
    """Class for Osipkov&Merrit stellar orbits See Mamon & Lokas 2005 for details."""

    def __init__(self):
        pass

    @staticmethod
    def K(r, R, r_ani):
        """Equation A16 im Mamon & Lokas 2005 for Osipkov&Merrit anisotropy.

        :param r: 3d radius
        :param R: projected 2d radius
        :param r_ani: anisotropy radius
        :return: K(r, R)
        """
        u = r / R
        ua = r_ani / R
        k = (ua**2 + 1.0 / 2) / (ua**2 + 1) ** (3.0 / 2) * (
            u**2 + ua**2
        ) / u * np.arctan(np.sqrt((u**2 - 1) / (ua**2 + 1))) - 1.0 / 2 / (
            ua**2 + 1
        ) * np.sqrt(
            1 - 1.0 / u**2
        )
        return k

    @staticmethod
    def beta_r(r, r_ani):
        """Anisotropy as a function of radius.

        :param r: 3d radius
        :param r_ani: anisotropy radius
        :return: beta
        """
        return r**2 / (r_ani**2 + r**2)

    @staticmethod
    def anisotropy_solution(r, r_ani):
        """The solution to d ln(f)/ d ln(r) = 2 beta(r) See e.g. A5 in Mamon & Lokas.

        :param r: 3d radius
        :param r_ani: anisotropy radius
        :return: f(r)
        """
        return r**2 + r_ani**2


@export
class GeneralizedOM(object):
    """Generalized Osipkov&Merrit profile.

    see Agnello et al. 2014 https://arxiv.org/pdf/1401.4462.pdf
    b(r) = beta_inf * r^2 / (r^2 + r_ani^2)
    """

    def __init__(self):
        self._z_interp = np.append(-np.flip(np.logspace(-1, 3, 200) ** 2), 0)
        # self._z_interp = -np.linspace(-200, 0, 200)**2  # z = (R**2 - r**2) / (r_ani**2 + R**2)

    @staticmethod
    def beta_r(r, r_ani, beta_inf):
        """Anisotropy as a function of radius.

        :param r: 3d radius
        :param r_ani: anisotropy radius
        :param beta_inf: anisotropy at infinity
        :return: beta
        """
        return beta_inf * r**2 / (r_ani**2 + r**2)

    def K(self, r, R, r_ani, beta_inf):
        """
        equation19 in Agnello et al. 2014 for k_beta(R, r) such that
        K(R, r) = (sqrt(r^2 - R^2) + k_beta(R, r)) / r

        :param r: 3d radius
        :param R: projected 2d radius
        :param r_ani: anisotropy radius
        :param beta_inf: anisotropy at infinity
        :return: K(r, R)
        """
        return (np.sqrt(r**2 - R**2) + self._k_beta(r, R, r_ani, beta_inf)) / r

    @staticmethod
    def anisotropy_solution(r, r_ani, beta_inf):
        """The solution to d ln(f)/ d ln(r) = 2 beta(r) See e.g. A5 in Mamon & Lokas
        with a scaling (nominator of Agnello et al. 2014 Equation (12)

        :param r: 3d radius
        :param r_ani: anisotropy radius
        :param beta_inf: anisotropy at infinity
        :return: f(r)
        """
        return (r**2 + r_ani**2) ** beta_inf

    def delete_cache(self):
        """Deletes the interpolation function of the hypergeometic function for a
        specific beta_inf.

        :return: deleted self variables
        """
        if hasattr(self, "_f_12_interp"):
            del self._f_12_interp
        if hasattr(self, "_f_32_interp"):
            del self._f_32_interp

    def _k_beta(self, r, R, r_ani, beta_inf):
        """
        equation19 in Agnello et al. 2014 for k_beta(R, r) such that
        K(R, r) = (sqrt(r^2 - R^2) + k_beta(R, r)) / r

        :param r: 3d radius
        :param R: projected 2d radius
        :param r_ani: anisotropy radius
        :param beta_inf: anisotropy at infinity
        :return: k_beta(r, R)
        """
        z = (R**2 - r**2) / (r_ani**2 + R**2)
        # ((r**2 + r_ani**2) / (R**2 + r_ani**2)) ** beta_inf
        return (
            -self.beta_r(R, r_ani, beta_inf)
            * self._j_beta(R, r, r_ani, beta_inf)
            * np.sqrt(r**2 - R**2)
            * (
                self._F_12(z, beta_inf)
                + 2.0 * (1 - r**2 / R**2) / 3 * self._F_32(z, beta_inf)
            )
        )

    def _F_12(self, z, beta_inf):
        """

        :param z: (R**2 - r**2) / (r_ani**2 + R**2)
        :param beta_inf: anisotropy at infinity
        :return: _F(1/2, z, beta_inf)
        """
        if not hasattr(self, "_f_12_interp"):
            f_12_interp = self._F(1 / 2.0, self._z_interp, beta_inf)
            self._f_12_interp = interp1d(
                self._z_interp, f_12_interp, kind="cubic", fill_value="extrapolate"
            )
        return self._f_12_interp(z)

    def _F_32(self, z, beta_inf):
        """

        :param z: (R**2 - r**2) / (r_ani**2 + R**2)
        :param beta_inf: anisotropy at infinity
        :return: _F(3/2, z, beta_inf)
        """
        if not hasattr(self, "_f_32_interp"):
            f_32_interp = self._F(3 / 2.0, self._z_interp, beta_inf)
            self._f_32_interp = interp1d(
                self._z_interp, f_32_interp, kind="cubic", fill_value="extrapolate"
            )
        return self._f_32_interp(z)

    @staticmethod
    def _j_beta(r, s, r_ani, beta_inf):
        """Equation (12) in Agnello et al. 2014.

        :param r:
        :param s:
        :param r_ani: :param beta_inf
        :return:
        """
        return ((s**2 + r_ani**2) / (r**2 + r_ani**2)) ** beta_inf

    @staticmethod
    def _F(a, z, beta_inf):
        """The hypergeometric function 2F1 (a, 1 +beta_inf, a + 1, z)

        :param a:
        :param z:
        :return:
        """
        if isinstance(z, int) or isinstance(z, float):
            return velocity_util.hyp_2F1(a=a, b=1 + beta_inf, c=a + 1, z=z)
        else:
            _F_array = []
            for z_i in z:
                _F_array.append(
                    velocity_util.hyp_2F1(a=a, b=1 + beta_inf, c=a + 1, z=z_i)
                )
            return np.array(_F_array, dtype=float)


@export
class Colin(object):
    """Class for stellar orbits anisotropy parameter based on Colin et al.

    (2000) See Mamon & Lokas 2005 for details
    """

    def __init__(self):
        pass

    @staticmethod
    def K(r, R, r_ani):
        """Equation A16 im Mamon & Lokas for Osipkov&Merrit anisotropy.

        :param r: 3d radius
        :param R: projected 2d radius
        :param r_ani: anisotropy radius
        :return: K(r, R)
        """
        u = r / R
        if np.min(u) < 1:
            raise ValueError(
                "3d radius is smaller than projected radius! Does not make sense."
            )
        ua = r_ani / R
        if ua == 1:
            k = (1 + 1.0 / u) * np.arccosh(u) - 1.0 / 6 * (8.0 / u + 7) * np.sqrt(
                (u - 1.0) / (u + 1.0)
            )
        elif ua > 1:
            k = (
                0.5 / (ua**2 - 1) * np.sqrt(1 - 1.0 / u**2)
                + (1.0 + ua / u) * np.arccosh(u)
                - np.sign(ua - 1)
                * ua
                * (ua**2 - 0.5)
                / np.abs(ua**2 - 1) ** (3.0 / 2)
                * (1.0 + ua / u)
                * np.arccosh((ua * u + 1) / (u + ua))
            )
        else:  # ua < 1
            k = (
                0.5 / (ua**2 - 1) * np.sqrt(1 - 1.0 / u**2)
                + (1.0 + ua / u) * np.arccosh(u)
                - np.sign(ua - 1)
                * ua
                * (ua**2 - 0.5)
                / np.abs(ua**2 - 1) ** (3.0 / 2)
                * (1.0 + ua / u)
                * np.arccos((ua * u + 1) / (u + ua))
            )
        return k

    @staticmethod
    def beta_r(r, r_ani):
        """Anisotropy as a function of radius.

        :param r: 3d radius
        :param r_ani: anisotropy radius
        :return: beta
        """
        return 1.0 / 2 * r / (r + r_ani)



__author__ = ["sibirrer", "ajshajib", "dgilman", "nataliehogg"]

import time

import numpy as np
from lenstronomy.Sampling.Samplers.pso import ParticleSwarmOptimizer
from lenstronomy.Util import sampling_util
from lenstronomy.Sampling.Pool.pool import choose_pool
from scipy.optimize import minimize

__all__ = ["Sampler"]


class Sampler(object):
    """
    class which executes the different sampling  methods
    Available are: affine-invariant ensemble sampling with emcee, ensemble slice sampling with zeus and a Particle Swarm Optimizer.
    These are examples and depending on your problem, you might find other/better solutions.
    Feel free to sample with your convenient sampler!

    """

    def __init__(self, likelihoodModule):
        """

        :param likelihoodModule: instance of LikelihoodModule class
        """
        self.chain = likelihoodModule
        self.lower_limit, self.upper_limit = self.chain.param_limits

    def simplex(self, init_pos, n_iterations, method, print_key="SIMPLEX"):
        """

        :param init_pos: starting point for the optimization
        :param n_iterations: maximum number of iterations
        :param method: the optimization method, default is 'Nelder-Mead'
        :return: the best fit for the lens model using the optimization routine specified by method
        """
        print("Performing the optimization using algorithm:", method)
        time_start = time.time()

        result = minimize(
            self.chain.negativelogL,
            x0=init_pos,
            method=method,
            options={"maxiter": n_iterations, "disp": True},
        )
        logL = self.chain.logL(result["x"])
        kwargs_return = self.chain.param.args2kwargs(result["x"])
        print(
            -logL * 2 / (max(self.chain.effective_num_data_points(**kwargs_return), 1)),
            "reduced X^2 of best position",
        )
        print(logL, "log likelihood")
        print(
            self.chain.effective_num_data_points(**kwargs_return),
            "effective number of data points",
        )
        print(kwargs_return.get("kwargs_lens", None), "lens result")
        print(kwargs_return.get("kwargs_source", None), "source result")
        print(kwargs_return.get("kwargs_lens_light", None), "lens light result")
        print(kwargs_return.get("kwargs_ps", None), "point source result")
        print(kwargs_return.get("kwargs_special", None), "special param result")
        time_end = time.time()
        print(time_end - time_start, "time used for ", print_key)
        print("===================")

        return result["x"]

    def pso(
        self,
        n_particles,
        n_iterations,
        lower_start=None,
        upper_start=None,
        threadCount=1,
        init_pos=None,
        mpi=False,
        print_key="PSO",
    ):
        """Return the best fit for the lens model on catalogue basis with particle swarm
        optimizer.

        :param n_particles: number of particles in the sampling process
        :param n_iterations: number of iterations of the swarm
        :param lower_start: numpy array, lower end parameter of the values of the
            starting particles
        :param upper_start: numpy array, upper end parameter of the values of the
            starting particles
        :param threadCount: number of threads in the computation (only applied if
            mpi=False)
        :param init_pos: numpy array, position of the initial best guess model
        :param mpi: bool, if True, makes instance of MPIPool to allow for MPI execution
        :param print_key: string, prints the process name in the progress bar (optional)
        :return: kwargs_result (of best fit), [lnlikelihood of samples, positions of
            samples, velocity of samples])
        """
        if lower_start is None or upper_start is None:
            lower_start, upper_start = np.array(self.lower_limit), np.array(
                self.upper_limit
            )
            print("PSO initialises its particles with default values")
        else:
            lower_start = np.maximum(lower_start, self.lower_limit)
            upper_start = np.minimum(upper_start, self.upper_limit)

        pool = choose_pool(mpi=mpi, processes=threadCount, use_dill=True)

        if mpi is True and pool.is_master():
            print("MPI option chosen for PSO.")

        pso = ParticleSwarmOptimizer(
            self.chain.logL, lower_start, upper_start, n_particles, pool=pool
        )

        if init_pos is None:
            init_pos = (upper_start - lower_start) / 2 + lower_start

        pso.set_global_best(init_pos, [0] * len(init_pos), self.chain.logL(init_pos))

        if pool.is_master():
            print("Computing the %s ..." % print_key)

        time_start = time.time()

        result, [log_likelihood_list, pos_list, vel_list] = pso.optimize(n_iterations)

        if pool.is_master():
            kwargs_return = self.chain.param.args2kwargs(result)
            print(
                pso.global_best.fitness
                * 2
                / (max(self.chain.effective_num_data_points(**kwargs_return), 1)),
                "reduced X^2 of best position",
            )
            print(pso.global_best.fitness, "log likelihood")
            print(
                self.chain.effective_num_data_points(**kwargs_return),
                "effective number of data points",
            )
            print(kwargs_return.get("kwargs_lens", None), "lens result")
            print(kwargs_return.get("kwargs_source", None), "source result")
            print(kwargs_return.get("kwargs_lens_light", None), "lens light result")
            print(kwargs_return.get("kwargs_ps", None), "point source result")
            print(kwargs_return.get("kwargs_special", None), "special param result")
            time_end = time.time()
            print(time_end - time_start, "time used for ", print_key)
            print("===================")
        return result, [log_likelihood_list, pos_list, vel_list]

    def mcmc_emcee(
        self,
        n_walkers,
        n_run,
        n_burn,
        mean_start,
        sigma_start,
        mpi=False,
        progress=False,
        threadCount=1,
        initpos=None,
        backend_filename=None,
        start_from_backend=False,
    ):
        """Run MCMC with emcee. For details, please have a look at the documentation of
        the emcee packager.

        :param n_walkers: number of walkers in the emcee process
        :type n_walkers: integer
        :param n_run: number of sampling (after burn-in) of the emcee
        :type n_run: integer
        :param n_burn: number of burn-in iterations (those will not be saved in the output sample)
        :type n_burn: integer
        :param mean_start: mean of the parameter position of the initialising sample
        :type mean_start: numpy array of length the number of parameters
        :param sigma_start: spread of the parameter values (uncorrelated in each dimension) of the initialising sample
        :type sigma_start: numpy array of length the number of parameters
        :param mpi: if True, initializes an MPIPool to allow for MPI execution of the sampler
        :type mpi: bool
        :param progress: if True, prints the progress bar
        :type progress: bool
        :param threadCount: number of threats in multi-processing (not applicable for MPI)
        :type threadCount: integer
        :param initpos: initial walker position to start sampling (optional)
        :type initpos: numpy array of size num param x num walkser
        :param backend_filename: name of the HDF5 file where sampling state is saved (through emcee backend engine)
        :type backend_filename: string
        :param start_from_backend: if True, start from the state saved in `backup_filename`.
         Otherwise, create a new backup file with name `backup_filename` (any already existing file is overwritten!).
        :type start_from_backend: bool
        :return: samples, ln likelihood value of samples
        :rtype: numpy 2d array, numpy 1d array
        """
        import emcee

        num_param, _ = self.chain.param.num_param()
        if initpos is None:
            initpos = sampling_util.sample_ball_truncated(
                mean_start,
                sigma_start,
                self.lower_limit,
                self.upper_limit,
                size=n_walkers,
            )

        pool = choose_pool(mpi=mpi, processes=threadCount, use_dill=True)

        if backend_filename is not None:
            backend = emcee.backends.HDFBackend(
                backend_filename, name="lenstronomy_mcmc_emcee"
            )
            if pool.is_master():
                print(
                    "Warning: All samples (including burn-in) will be saved in backup file '{}'.".format(
                        backend_filename
                    )
                )
            if start_from_backend:
                initpos = None
                n_run_eff = n_run
            else:
                n_run_eff = n_burn + n_run
                backend.reset(n_walkers, num_param)
                if pool.is_master():
                    print(
                        "Warning: backup file '{}' has been reset!".format(
                            backend_filename
                        )
                    )
        else:
            backend = None
            n_run_eff = n_burn + n_run

        time_start = time.time()

        sampler = emcee.EnsembleSampler(
            n_walkers, num_param, self.chain.logL, pool=pool, backend=backend
        )

        sampler.run_mcmc(initpos, n_run_eff, progress=progress)
        flat_samples = sampler.get_chain(discard=n_burn, thin=1, flat=True)
        dist = sampler.get_log_prob(flat=True, discard=n_burn, thin=1)
        if pool.is_master():
            print("Computing the MCMC...")
            print("Number of walkers = ", n_walkers)
            print("Burn-in iterations: ", n_burn)
            print("Sampling iterations (in current run):", n_run_eff)
            time_end = time.time()
            print(time_end - time_start, "time taken for MCMC sampling")
        return flat_samples, dist

    def mcmc_zeus(
        self,
        n_walkers,
        n_run,
        n_burn,
        mean_start,
        sigma_start,
        mpi=False,
        threadCount=1,
        progress=False,
        initpos=None,
        backend_filename=None,
        **kwargs_zeus
    ):
        """
        Lightning fast MCMC with zeus: https://github.com/minaskar/zeus

        For the full list of arguments for the EnsembleSampler and callbacks, see see `the zeus docs <https://zeus-mcmc.readthedocs.io/en/latest/api/sampler.html>`_.

        If you use the zeus sampler, you should cite the following papers: 2105.03468, 2002.06212.

        :param n_walkers: number of walkers per parameter
        :type n_walkers: integer
        :param n_run: number of sampling steps
        :type n_run: integer
        :param n_burn: number of burn-in steps
        :type n_burn: integer
        :param mean_start: mean of the parameter position of the initialising sample
        :type mean_start: numpy array of length the number of parameters
        :param sigma_start: spread of the parameter values (uncorrelated in each dimension) of the initialising sample
        :type sigma_start: numpy array of length the number of parameters
        :param mpi: if True, initializes an MPIPool to allow for MPI execution of the sampler
        :type mpi: bool
        :param progress:
        :type progress: bool
        :param initpos: initial walker position to start sampling (optional)
        :type initpos: numpy array of size num param x num walkser
        :param backend_filename: name of the HDF5 file where sampling state is saved (through zeus callback function)
        :type backend_filename: string
        :return: samples, ln likelihood value of samples
        :rtype: numpy 2d array, numpy 1d array
        """
        import zeus

        print("Using zeus to perform the MCMC.")

        num_param, _ = self.chain.param.num_param()

        # zeus kwargs; checks the dict for the key and if not present returns the given value
        moves = kwargs_zeus.get("moves")
        tune = kwargs_zeus.get("tune", True)
        tolerance = kwargs_zeus.get("tolerance", 0.05)
        patience = kwargs_zeus.get("patience", 5)
        maxsteps = kwargs_zeus.get("maxsteps", 10000)
        mu = kwargs_zeus.get("mu", 1.0)
        maxiter = kwargs_zeus.get("maxiter", 10000)
        pool = kwargs_zeus.get("pool", None)
        vectorize = kwargs_zeus.get("vectorize", False)
        blobs_dtype = kwargs_zeus.get("blobs_dtype")
        verbose = kwargs_zeus.get("verbose", True)
        check_walkers = kwargs_zeus.get("check_walkers", True)
        shuffle_ensemble = kwargs_zeus.get("shuffle_ensemble", True)
        light_mode = kwargs_zeus.get("light_mode", False)

        # kwargs specifically for the callbacks
        autocorrelation_callback = kwargs_zeus.get("autocorrelation_callback", False)
        ncheck = kwargs_zeus.get("ncheck", 1)
        dact = kwargs_zeus.get("dact", 0.01)
        nact = kwargs_zeus.get("nact", 50)
        discard = kwargs_zeus.get("discard", 0.5)
        trigger = kwargs_zeus.get("trigger", True)
        method = kwargs_zeus.get("method", "mk")
        splitr_callback = kwargs_zeus.get("splitr_callback", False)
        epsilon = kwargs_zeus.get("epsilon", 0.01)
        nsplits = kwargs_zeus.get("nsplits", 2)
        miniter_callback = kwargs_zeus.get("miniter_callback", False)
        nmin = kwargs_zeus.get("nmin", 1000)

        if initpos is None:
            initpos = sampling_util.sample_ball_truncated(
                mean_start,
                sigma_start,
                self.lower_limit,
                self.upper_limit,
                size=n_walkers,
            )

        n_run_eff = n_burn + n_run

        callback_list = []

        if backend_filename is not None:
            backend = zeus.callbacks.SaveProgressCallback(
                filename=backend_filename, ncheck=ncheck
            )
            callback_list.append(backend)
        else:
            pass

        if autocorrelation_callback == True:
            autocorrelation = zeus.callbacks.AutocorrelationCallback(
                ncheck=ncheck,
                dact=dact,
                nact=nact,
                discard=discard,
                trigger=trigger,
                method=method,
            )
            callback_list.append(autocorrelation)
        else:
            pass

        if splitr_callback == True:
            splitr = zeus.callbacks.SplitRCallback(
                ncheck=ncheck,
                epsilon=epsilon,
                nsplits=nsplits,
                discard=discard,
                trigger=trigger,
            )
            callback_list.append(splitr)
        else:
            pass

        if miniter_callback == True:
            miniter = zeus.callbacks.MinIterCallback(nmin=nmin)
            callback_list.append(miniter)
        else:
            pass

        pool = choose_pool(mpi=mpi, processes=threadCount, use_dill=True)

        sampler = zeus.EnsembleSampler(
            nwalkers=n_walkers,
            ndim=num_param,
            logprob_fn=self.chain.logL,
            moves=moves,
            tune=tune,
            tolerance=tolerance,
            patience=patience,
            maxsteps=maxsteps,
            mu=mu,
            maxiter=maxiter,
            pool=pool,
            vectorize=vectorize,
            blobs_dtype=blobs_dtype,
            verbose=verbose,
            check_walkers=check_walkers,
            shuffle_ensemble=shuffle_ensemble,
            light_mode=light_mode,
        )

        sampler.run_mcmc(initpos, n_run_eff, progress=progress, callbacks=callback_list)

        flat_samples = sampler.get_chain(flat=True, thin=1, discard=n_burn)

        dist = sampler.get_log_prob(flat=True, thin=1, discard=n_burn)

        return flat_samples, dist



from lenstronomy.LightModel.linear_basis import LinearBasis


class TestLinearBasis(object):
    def setup_method(self):
        pass

    def test_linear_param_from_kwargs(self):
        linear_basis = LinearBasis(light_model_list=["UNIFORM", "UNIFORM"])
        kwargs_list = [{"amp": 0.5}, {"amp": -1}]
        param = linear_basis.linear_param_from_kwargs(kwargs_list)
        assert param[0] == kwargs_list[0]["amp"]
        assert param[1] == kwargs_list[1]["amp"]



import numpy as np
import numpy.testing as npt
from lenstronomy.Data.imaging_data import ImageData


def test_interferometry_likelihood():
    """Test interferometry natural weighting likelihood function, test likelihood_method
    function output."""

    test_data = np.zeros((5, 5))
    test_data[0, :] = 1
    test_data[:, 4] = 2

    mask = np.ones((5, 5))
    model_unconvolved = np.zeros((5, 5))
    model_convolved = np.zeros((5, 5))
    model_unconvolved[0, 4] = 1.2
    model_convolved[0, :] = 1
    model_convolved[:, 4] = 1

    data_class = ImageData(
        image_data=test_data,
        background_rms=2.5,
        log_likelihood_constant=-1.0,
        likelihood_method="interferometry_natwt",
    )

    assert data_class.likelihood_method() == "interferometry_natwt"
    npt.assert_almost_equal(
        data_class.log_likelihood([model_unconvolved, model_convolved], mask),
        -0.712,
        decimal=8,
    )



import pytest
import numpy.testing as npt
import numpy as np
from lenstronomy.LightModel.Profiles.nie import NIE as NIE_light
from lenstronomy.LensModel.Profiles.nie import NIE as NIE_lens
from lenstronomy.Util import util


class TestNIE(object):
    """Class to test the NIE profile."""

    def setup_method(self):
        pass

    def test_function(self):
        """

        :return:
        """
        lens = NIE_lens()
        light = NIE_light()

        x, y = util.make_grid(numPix=100, deltapix=0.1)
        e1, e2 = 0.2, 0
        s = 1.0
        kwargs_light = {"amp": 1.0, "e1": e1, "e2": e2, "s_scale": s}
        kwargs_lens = {"theta_E": 1.0, "e1": e1, "e2": e2, "s_scale": s}
        flux = light.function(x=x, y=y, **kwargs_light)
        f_xx, f_xy, f_yx, f_yy = lens.hessian(x=x, y=y, **kwargs_lens)
        kappa = 1 / 2.0 * (f_xx + f_yy)

        npt.assert_almost_equal(flux / flux[-1], kappa / kappa[-1], decimal=3)

        # test whether ellipticity changes overall flux normalization
        kwargs_light_round = {"amp": 1.0, "e1": 0, "e2": 0, "s_scale": s}

        x_, y_ = util.points_on_circle(radius=1, num_points=20)
        f_r = light.function(x=x_, y=y_, **kwargs_light)
        f_r = np.mean(f_r)

        f_r_round = light.function(x=1, y=0, **kwargs_light_round)
        npt.assert_almost_equal(f_r / f_r_round, 1, decimal=2)


if __name__ == "__main__":
    pytest.main()



import numpy.testing as npt
import numpy as np

from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.ImSim.image_model import ImageModel
import lenstronomy.Util.simulation_util as sim_util
from lenstronomy.Util import kernel_util
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF
import scipy.signal

"""
test the antenna primary beam and interferometric PSF (containing negative pixels and do not want normalisation)
The idea of this test is to define two sets of data class and psf class, one with the antenna primary beam and PSF, one without, 
and compare the (image_with_pb_and_psf) with scipy.signal.fftconvolve(image_without_pb_psf * pb, PSF, mode='same').
"""


def test_ImageModel_with_primary_beam_and_interferometry_psf():
    sigma_bkg = 0.05
    exp_time = np.inf
    numPix = 100
    deltaPix = 0.05

    # simulate a primary beam (pb)
    primary_beam = np.zeros((numPix, numPix))
    for i in range(numPix):
        for j in range(numPix):
            primary_beam[i, j] = np.exp(-1e-4 * ((i - 78) ** 2 + (j - 56) ** 2))
    primary_beam /= np.max(primary_beam)

    # simulate a spherical sinc function as psf, which contains negative pixels
    psf_test = np.zeros((221, 221))
    for i in range(221):
        for j in range(221):
            if i > j:
                psf_test[i, j] = psf_test[j, i]
            r = np.sqrt((i - 110) ** 2 + (j - 110) ** 2)
            if r == 0:
                psf_test[i, j] = 1
            else:
                psf_test[i, j] = np.sin(r * 0.5) / (r * 0.5)

    # define two data classes
    kwargs_data_no_pb = sim_util.data_configure_simple(
        numPix, deltaPix, exp_time, sigma_bkg
    )
    data_class_no_pb = ImageData(**kwargs_data_no_pb)

    kwargs_data_with_pb = sim_util.data_configure_simple(
        numPix, deltaPix, exp_time, sigma_bkg
    )
    kwargs_data_with_pb["antenna_primary_beam"] = primary_beam
    data_class_with_pb = ImageData(**kwargs_data_with_pb)

    # define two psf classes
    kwargs_psf_none = {"psf_type": "NONE"}
    psf_class_none = PSF(**kwargs_psf_none)

    kernel_cut = kernel_util.cut_psf(psf_test, 201, normalisation=False)
    kwargs_psf = {
        "psf_type": "PIXEL",
        "pixel_size": deltaPix,
        "kernel_point_source": kernel_cut,
        "kernel_point_source_normalisation": False,
    }
    psf_class = PSF(**kwargs_psf)

    # define lens model and source model
    kwargs_shear = {"gamma1": 0.01, "gamma2": 0.01}
    kwargs_spemd = {
        "theta_E": 1.0,
        "gamma": 1.8,
        "center_x": 0,
        "center_y": 0,
        "e1": 0.1,
        "e2": 0.04,
    }
    lens_model_list = ["SPEP", "SHEAR"]
    kwargs_lens = [kwargs_spemd, kwargs_shear]
    lens_model_class = LensModel(lens_model_list=lens_model_list)

    kwargs_sersic = {
        "amp": 30.0,
        "R_sersic": 0.3,
        "n_sersic": 2,
        "center_x": 0,
        "center_y": 0,
    }
    lens_light_model_list = ["SERSIC"]
    kwargs_lens_light = [kwargs_sersic]
    lens_light_model_class = LightModel(light_model_list=lens_light_model_list)

    kwargs_sersic_ellipse = {
        "amp": 1.0,
        "R_sersic": 0.6,
        "n_sersic": 7,
        "center_x": 0,
        "center_y": 0,
        "e1": 0.05,
        "e2": 0.02,
    }
    source_model_list = ["SERSIC_ELLIPSE"]
    kwargs_source = [kwargs_sersic_ellipse]
    source_model_class = LightModel(light_model_list=source_model_list)

    kwargs_numerics = {"supersampling_factor": 1, "supersampling_convolution": False}

    # make images using 1) data and psf classes without pb and or psf
    imageModel_no_pb_psf = ImageModel(
        data_class_no_pb,
        psf_class_none,
        lens_model_class,
        source_model_class,
        lens_light_model_class,
        kwargs_numerics=kwargs_numerics,
    )
    image_sim_no_pb_psf = imageModel_no_pb_psf.image(
        kwargs_lens, kwargs_source, kwargs_lens_light
    )

    # make images using 2) data and psf classes with defined pb and or psf
    imageModel_with_pb_psf = ImageModel(
        data_class_with_pb,
        psf_class,
        lens_model_class,
        source_model_class,
        lens_light_model_class,
        kwargs_numerics=kwargs_numerics,
    )
    image_sim_with_pb_psf = imageModel_with_pb_psf.image(
        kwargs_lens, kwargs_source, kwargs_lens_light
    )

    # add pb and psf to 1) out of the imageModel, compare them to check if the pb and psf changes make sense
    image_sim_with_pb_psf2 = scipy.signal.fftconvolve(
        image_sim_no_pb_psf * primary_beam, kernel_cut, mode="same"
    )
    npt.assert_almost_equal(image_sim_with_pb_psf, image_sim_with_pb_psf2, decimal=8)



import numpy as np
import warnings
import lenstronomy.Util.data_util as data_util
from lenstronomy.Data.psf import PSF

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
class Instrument(object):
    """Basic access points to instrument properties."""

    def __init__(self, pixel_scale, read_noise=None, ccd_gain=None):
        """

        :param read_noise: std of noise generated by read-out (in units of electrons)
        :param pixel_scale: scale (in arcseconds) of pixels
        :param ccd_gain: electrons/ADU (analog-to-digital unit).
         A gain of 8 means that the camera digitizes the CCD signal
         so that each ADU corresponds to 8 photoelectrons.
        """
        if ccd_gain is not None:
            ccd_gain = float(ccd_gain)
        self.ccd_gain = ccd_gain
        self._read_noise = read_noise
        self.pixel_scale = pixel_scale


@export
class Observation(object):
    """Basic access point to observation properties."""

    def __init__(
        self,
        exposure_time,
        sky_brightness=None,
        seeing=None,
        num_exposures=1,
        psf_type="GAUSSIAN",
        kernel_point_source=None,
        truncation=5,
        point_source_supersampling_factor=1,
    ):
        """

        :param exposure_time: exposure time per image (in seconds)
        :param sky_brightness: sky brightness (in magnitude per square arcseconds)
        :param seeing: full width at half maximum of the PSF (if not specific psf_model is specified)
        :param num_exposures: number of exposures that are combined
        :param psf_type: string, type of PSF ('GAUSSIAN' and 'PIXEL' supported)
        :param kernel_point_source: 2d numpy array, model of PSF centered with odd number of pixels per axis
         (optional when psf_type='PIXEL' is chosen)
        :param point_source_supersampling_factor: int, supersampling factor of kernel_point_source
         (optional when psf_type='PIXEL' is chosen)
        """
        self._exposure_time = exposure_time
        self._sky_brightness_ = sky_brightness
        self._num_exposures = num_exposures
        self._seeing = seeing
        self._psf_type = psf_type
        self._truncation = truncation
        self._kernel_point_source = kernel_point_source
        self._point_source_supersampling_factor = point_source_supersampling_factor

    def update_observation(
        self,
        exposure_time=None,
        sky_brightness=None,
        seeing=None,
        num_exposures=None,
        psf_type=None,
        kernel_point_source=None,
    ):
        """Updates class instance with new properties if specific argument is not None.

        :param exposure_time: exposure time per image (in seconds)
        :param sky_brightness: sky brightness (in magnitude per square arcseconds)
        :param seeing: full width at half maximum of the PSF (if not specific psf_model
            is specified)
        :param num_exposures: number of exposures that are combined
        :param psf_type: string, type of PSF ('GAUSSIAN' and 'PIXEL' supported)
        :param kernel_point_source: 2d numpy array, model of PSF centered with odd
            number of pixels per axis (optional when psf_type='PIXEL' is chosen)
        :return: None, updated class instance
        """
        if exposure_time is not None:
            self._exposure_time = exposure_time
        if sky_brightness is not None:
            self._sky_brightness_ = sky_brightness
        if seeing is not None:
            self._seeing = seeing
        if num_exposures is not None:
            self._num_exposures = num_exposures
        if psf_type is not None:
            self._psf_type = psf_type
        if kernel_point_source is not None:
            self._kernel_point_source = kernel_point_source

    @property
    def _sky_brightness(self):
        if self._sky_brightness_ is None:
            raise ValueError("sky_brightness is not set in the class instance!")
        return self._sky_brightness_

    @property
    def exposure_time(self):
        """Total exposure time.

        :return: summed exposure time
        """
        return self._exposure_time * self._num_exposures

    @property
    def kwargs_psf(self):
        """Keyword arguments to initiate a PSF() class.

        :return: kwargs_psf
        """
        if self._psf_type == "GAUSSIAN":
            psf_type = "GAUSSIAN"
            fwhm = self._seeing
            truncation = self._truncation
            kwargs_psf = {"psf_type": psf_type, "fwhm": fwhm, "truncation": truncation}
        elif self._psf_type == "PIXEL":
            if self._kernel_point_source is not None:
                kwargs_psf = {
                    "psf_type": "PIXEL",
                    "kernel_point_source": self._kernel_point_source,
                    "point_source_supersampling_factor": self._point_source_supersampling_factor,
                }
            else:
                raise ValueError(
                    "You need to create the class instance with a psf_model!"
                )
        elif self._psf_type == "NONE":
            kwargs_psf = {"psf_type": "NONE"}
        else:
            raise ValueError("psf_type %s not supported!" % self._psf_type)
        return kwargs_psf

    @property
    def psf_class(self):
        """Creates instance of PSF() class based on knowledge of the observations For
        the full possibility of how to create such an instance, see the PSF() class
        documentation.

        :return: instance of PSF() class
        """
        psf_class = PSF(**self.kwargs_psf)
        return psf_class


@export
class SingleBand(Instrument, Observation):
    """Class that combines Instrument and Observation."""

    def __init__(
        self,
        pixel_scale,
        exposure_time,
        magnitude_zero_point,
        read_noise=None,
        ccd_gain=None,
        sky_brightness=None,
        seeing=None,
        num_exposures=1,
        psf_type="GAUSSIAN",
        kernel_point_source=None,
        truncation=5,
        point_source_supersampling_factor=1,
        data_count_unit="e-",
        background_noise=None,
    ):
        """

        :param read_noise: std of noise generated by read-out (in units of electrons)
        :param pixel_scale: scale (in arcseconds) of pixels
        :param ccd_gain: electrons/ADU (analog-to-digital unit).
         A gain of 8 means that the camera digitizes the CCD signal
         so that each ADU corresponds to 8 photoelectrons.
        :param exposure_time: exposure time per image (in seconds)
        :param sky_brightness: sky brightness (in magnitude per square arcseconds in units of electrons)
        :param seeing: Full-Width-at-Half-Maximum (FWHM) of PSF
        :param magnitude_zero_point: magnitude in which 1 count (e-) per second per arcsecond square is registered
        :param num_exposures: number of exposures that are combined
        :param point_source_supersampling_factor: int, supersampling factor of kernel_point_source
         (optional when psf_type='PIXEL' is chosen)
        :param data_count_unit: string, unit of the data (not noise properties - see other definitions),
         'e-': (electrons assumed to be IID),
         'ADU': (analog-to-digital unit)
        :param background_noise: sqrt(variance of background) as a total contribution from readnoise,
         sky brightness etc. in units of the data_count_units (e- or ADU)
         If you set this parameter, it will use this value regardless of the values of read_noise, sky_brightness
        """
        Instrument.__init__(
            self, pixel_scale, read_noise, ccd_gain
        )  # read_noise and ccd_gain can be None
        Observation.__init__(
            self,
            exposure_time=exposure_time,
            sky_brightness=sky_brightness,
            seeing=seeing,
            num_exposures=num_exposures,
            psf_type=psf_type,
            kernel_point_source=kernel_point_source,
            point_source_supersampling_factor=point_source_supersampling_factor,
            truncation=truncation,
        )
        if data_count_unit not in ["e-", "ADU"]:
            raise ValueError(
                "count_unit type %s not supported! Please choose e- or ADU."
                % data_count_unit
            )
        self._data_count_unit = data_count_unit
        self._background_noise = background_noise
        self._magnitude_zero_point = magnitude_zero_point

    @property
    def sky_brightness(self):
        """

        :return: sky brightness (counts per square arcseconds in unit of data (e- or ADU's) per unit time)
        """
        cps = self._sky_brightness_cps
        if self._data_count_unit == "ADU":
            cps /= self.ccd_gain
        return cps

    @property
    def _sky_brightness_cps(self):
        """

        :return: sky brightness in electrons per second
        """
        cps = data_util.magnitude2cps(
            self._sky_brightness, magnitude_zero_point=self._magnitude_zero_point
        )
        return cps

    @property
    def background_noise(self):
        """Gaussian sigma of noise level per pixel in counts (e- or ADU) per second.

        :return: sqrt(variance) of background noise level in data units
        """
        if self._background_noise is None:
            if self._read_noise is None:
                raise ValueError(
                    "read_noise is not specified to evaluate background noise!"
                )
            bkg_noise = data_util.bkg_noise(
                self._read_noise,
                self._exposure_time,
                self._sky_brightness_cps,
                self.pixel_scale,
                num_exposures=self._num_exposures,
            )
            if self._data_count_unit == "ADU":
                bkg_noise /= self.ccd_gain
            return bkg_noise
        else:
            if self._read_noise is not None:
                warnings.warn(
                    "read noise is specified but not used for noise properties. Background noise is estimated"
                    ' from "background_noise" argument.'
                )
            return self._background_noise

    def flux_noise(self, flux):
        """

        :param flux: float or array, units of count_unit/seconds, needs to be positive semi-definite in the flux value
        :return: Gaussian approximation of Poisson statistics in IIDs sqrt(variance)
        """
        flux_iid = self.flux_iid(flux)
        variance = (
            flux_iid  # the variance of a Poisson distribution is the IID count number
        )
        if isinstance(variance, int) or isinstance(variance, float):
            variance = max(variance, 0)
        else:
            variance[
                flux_iid < 0
            ] = 0  # make sure negative pixels do not lead to variances (or nans) in the return
        noise = np.sqrt(variance) / self.exposure_time
        if self._data_count_unit == "ADU":
            noise /= self.ccd_gain
        return noise

    def flux_iid(self, flux_per_second):
        """IID counts. This can be used by lenstronomy to estimate the Poisson errors
        keeping the assumption that the counts are IIDs (even if they are not).

        :param flux_per_second: flux count per second in the units set in this class
            (ADU or e-)
        :return: IID count number
        """
        if self._data_count_unit == "ADU":
            exp_time = self.ccd_gain * self.exposure_time
        else:
            exp_time = self.exposure_time
        return exp_time * flux_per_second

    def noise_for_model(
        self, model, background_noise=True, poisson_noise=True, seed=None
    ):
        """

        :param model: 2d numpy array of modelled image (with pixels in units of data specified in class)
        :param background_noise: bool, if True, adds background noise
        :param poisson_noise: bool, if True, adds Poisson noise of modelled flux
        :param seed: int, seed number to be used to render the noise properties.
         If None, then uses the current numpy.random seed to render the noise properties.
        :return: noise realization corresponding to the model
        """
        if seed is not None:
            g = np.random.RandomState(seed=seed)
        else:
            g = np.random
        nx, ny = np.shape(model)
        noise = np.zeros_like(model)
        if background_noise is True:
            noise += g.randn(nx, ny) * self.background_noise
        if poisson_noise is True:
            noise += g.randn(nx, ny) * self.flux_noise(model)
        return noise

    def estimate_noise(self, image):
        """

        :param image: noisy data, background subtracted
        :return: estimated noise map  sqrt(variance) for each pixel as estimated from the instrument and observation
        """
        return np.sqrt(self.background_noise**2 + self.flux_noise(image) ** 2)

    def magnitude2cps(self, magnitude):
        """Converts an apparent magnitude to counts per second (in units of the data)

        The zero point of an instrument, by definition, is the magnitude of an object that produces one count
        (or data number, DN) per second. The magnitude of an arbitrary object producing DN counts in an observation of
        length EXPTIME is therefore:
        m = -2.5 x log10(DN / EXPTIME) + ZEROPOINT

        :param magnitude: magnitude of object
        :return: counts per second of object
        """
        # compute counts in units of e- or ADS (depending on data and magnitude zero point defined)
        cps = data_util.magnitude2cps(
            magnitude, magnitude_zero_point=self._magnitude_zero_point
        )
        if self._data_count_unit == "ADU":
            cps /= self.ccd_gain
        return cps



import yaml
import os
import lenstronomy


# in case the xdg library is installed, the import statement with pyxdg can raise an error
# to avoid it, we draw back to the ~/.config directory in case this import fails.
# TODO come up with more permanent solution of path to configuration directory
try:
    from xdg.BaseDirectory import xdg_config_home
except ImportError:
    xdg_config_home = "~/.config"

user_config_file = os.path.join(xdg_config_home, "lenstronomy", "config.yaml")

module_path = os.path.dirname(lenstronomy.__file__)
default_config_file = os.path.join(module_path, "Conf", "conf_default.yaml")

if os.path.exists(user_config_file):
    conf_file = user_config_file
else:
    conf_file = default_config_file

with open(conf_file) as file:
    # The FullLoader parameter handles the conversion from YAML
    # scalar values to the Python the dictionary format
    conf = yaml.safe_load(file)
    # conf = yaml.load(file, Loader=yaml.FullLoader)
    numba_conf = conf["numba"]
    nopython = numba_conf["nopython"]
    cache = numba_conf["cache"]
    parallel = numba_conf["parallel"]
    numba_enabled = numba_conf["enable"]
    fastmath = numba_conf["fastmath"]
    error_model = numba_conf["error_model"]


def numba_conf():
    """

    :return: keyword arguments of numba configurations from yaml file
    """
    with open(conf_file) as file:
        # The FullLoader parameter handles the conversion from YAML
        # scalar values to the Python the dictionary format
        conf = yaml.safe_load(file)
        # conf = yaml.load(file, Loader=yaml.FullLoader)
        numba_conf = conf["numba"]
    return numba_conf


def conventions_conf():
    """

    :return: convention keyword arguments
    """
    with open(conf_file) as file:
        conf = yaml.safe_load(file)
        conventions_conf = conf["conventions"]
    return conventions_conf



from lenstronomy.SimulationAPI.sim_api import SimAPI
import lenstronomy.SimulationAPI.observation_constructor as constructor
import pytest
import numpy.testing as npt


class TestModelAPI(object):
    def setup_method(self):
        numpix = 10
        instrument_name = "LSST"
        observation_name = "LSST_g_band"
        kwargs_single_band = constructor.observation_constructor(
            instrument_name=instrument_name, observation_name=observation_name
        )
        kwargs_single_band["data_count_unit"] = "e-"
        kwargs_model = {
            "lens_model_list": ["SIS"],
            "z_lens": None,
            "z_source": None,
            "lens_redshift_list": None,
            "source_light_model_list": ["GAUSSIAN"],
            "lens_light_model_list": ["SERSIC"],
            "point_source_model_list": ["UNLENSED"],
            "source_redshift_list": None,
        }
        kwargs_numerics = {"supersampling_factor": 2}
        self.api = SimAPI(numpix, kwargs_single_band, kwargs_model)

    def test_image_model_class(self):
        model = self.api.image_model_class()
        assert model.LensModel.lens_model_list[0] == "SIS"

    def test_magnitude2amplitude(self):
        kwargs_lens_light_mag = [
            {
                "magnitude": 28,
                "R_sersic": 1.0,
                "n_sersic": 2,
                "center_x": 0,
                "center_y": 0,
            }
        ]
        kwargs_source_mag = [
            {"magnitude": 30, "sigma": 0.3, "center_x": 0, "center_y": 0}
        ]
        kwargs_ps_mag = [{"magnitude": [30], "ra_image": [0], "dec_image": [0]}]
        kwargs_lens_light, kwargs_source, kwargs_ps = self.api.magnitude2amplitude(
            kwargs_lens_light_mag, kwargs_source_mag, kwargs_ps_mag
        )

        npt.assert_almost_equal(kwargs_source[0]["amp"], 1, decimal=5)
        npt.assert_almost_equal(kwargs_ps[0]["point_amp"][0], 1, decimal=5)
        npt.assert_almost_equal(
            kwargs_lens_light[0]["amp"], 0.38680586575451237, decimal=5
        )


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import numpy.testing as npt
import numpy as np
import pytest
import unittest
from lenstronomy.LensModel.MultiPlane.multi_plane import MultiPlane
from lenstronomy.LensModel.MultiPlane.multi_plane_base import MultiPlaneBase
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.MultiPlane.multi_plane import (
    LensedLocation,
    PhysicalLocation,
)
import lenstronomy.Util.constants as const


class TestMultiPlane(object):
    """Tests the source model routines."""

    def setup_method(self):
        pass

    def test_update_source_redshift(self):
        z_source = 1.5
        lens_model_list = ["SIS"]
        kwargs_lens = [{"theta_E": 1}]
        redshift_list = [0.5]
        lensModelMutli = MultiPlane(
            z_source=z_source,
            lens_model_list=lens_model_list,
            lens_redshift_list=redshift_list,
            z_interp_stop=3,
            cosmo_interp=True,
        )
        alpha_x, alpha_y = lensModelMutli.alpha(1, 0, kwargs_lens=kwargs_lens)
        lensModelMutli.update_source_redshift(z_source=z_source)
        alpha_x_new, alpha_y_new = lensModelMutli.alpha(1, 0, kwargs_lens=kwargs_lens)
        npt.assert_almost_equal(alpha_x / alpha_x_new, 1.0, decimal=8)

        lensModelMutli.update_source_redshift(z_source=1.0)
        alpha_x_new, alpha_y_new = lensModelMutli.alpha(1, 0, kwargs_lens=kwargs_lens)
        assert alpha_x / alpha_x_new > 1

        lensModelMutli.update_source_redshift(z_source=2.0)
        alpha_x_new, alpha_y_new = lensModelMutli.alpha(1, 0, kwargs_lens=kwargs_lens)
        assert alpha_x / alpha_x_new < 1

    def test_sis_alpha(self):
        z_source = 1.5
        lens_model_list = ["SIS"]
        redshift_list = [0.5]
        lensModelMutli = MultiPlane(
            z_source=z_source,
            lens_model_list=lens_model_list,
            lens_redshift_list=redshift_list,
        )
        lensModel = LensModel(lens_model_list=lens_model_list)
        kwargs_lens = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
        alpha_x_simple, alpha_y_simple = lensModel.alpha(1, 0, kwargs_lens)
        alpha_x_multi, alpha_y_multi = lensModelMutli.alpha(1, 0, kwargs_lens)
        npt.assert_almost_equal(alpha_x_simple, alpha_x_multi, decimal=8)
        npt.assert_almost_equal(alpha_y_simple, alpha_y_multi, decimal=8)
        sum_partial = (
            np.sum(lensModelMutli._multi_plane_base._T_ij_list)
            + lensModelMutli._T_ij_stop
        )
        T_z_true = lensModelMutli._T_z_source
        npt.assert_almost_equal(sum_partial, T_z_true, decimal=5)

    def test_sis_ray_tracing(self):
        z_source = 1.5
        lens_model_list = ["SIS"]
        redshift_list = [0.5]
        from astropy.cosmology import FlatLambdaCDM, LambdaCDM

        # test flat LCDM
        cosmo = FlatLambdaCDM(H0=70, Om0=0.3)
        lensModelMutli = MultiPlane(
            z_source=z_source,
            lens_model_list=lens_model_list,
            lens_redshift_list=redshift_list,
            cosmo=cosmo,
        )
        lensModel = LensModel(lens_model_list=lens_model_list, cosmo=cosmo)
        kwargs_lens = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
        beta_x_simple, beta_y_simple = lensModel.ray_shooting(1, 0, kwargs_lens)
        beta_x_multi, beta_y_multi = lensModelMutli.ray_shooting(1, 0, kwargs_lens)
        npt.assert_almost_equal(beta_x_simple, beta_x_multi, decimal=10)
        npt.assert_almost_equal(beta_y_simple, beta_y_multi, decimal=10)
        npt.assert_almost_equal(beta_x_simple, 0, decimal=10)
        npt.assert_almost_equal(beta_y_simple, 0, decimal=10)

    def test_sis_hessian(self):
        z_source = 1.5
        lens_model_list = ["SIS"]
        redshift_list = [0.5]
        lensModelMutli = MultiPlane(
            z_source=z_source,
            lens_model_list=lens_model_list,
            lens_redshift_list=redshift_list,
        )
        lensModel = LensModel(lens_model_list=lens_model_list)
        kwargs_lens = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
        f_xx_simple, f_xy_simple, f_yx_simple, f_yy_simple = lensModel.hessian(
            1, 0, kwargs_lens
        )
        f_xx_multi, f_xy_multi, f_yx_multi, f_yy_multi = lensModelMutli.hessian(
            1, 0, kwargs_lens, diff=0.000001
        )
        npt.assert_almost_equal(f_xx_simple, f_xx_multi, decimal=5)
        npt.assert_almost_equal(f_xy_simple, f_xy_multi, decimal=5)
        npt.assert_almost_equal(f_yx_simple, f_yx_multi, decimal=5)
        npt.assert_almost_equal(f_yy_simple, f_yy_multi, decimal=5)

    def test_empty(self):
        z_source = 1.5
        lens_model_list = []
        redshift_list = []
        lensModelMutli = MultiPlane(
            z_source=z_source,
            lens_model_list=lens_model_list,
            lens_redshift_list=redshift_list,
        )
        kwargs_lens = []
        f_xx_multi, f_xy_multi, f_yx_multi, f_yy_multi = lensModelMutli.hessian(
            1, 0, kwargs_lens, diff=0.000001
        )
        npt.assert_almost_equal(0, f_xx_multi, decimal=5)
        npt.assert_almost_equal(0, f_xy_multi, decimal=5)
        npt.assert_almost_equal(0, f_yx_multi, decimal=5)
        npt.assert_almost_equal(0, f_yy_multi, decimal=5)

    def test_sis_kappa_gamma_mag(self):
        z_source = 1.5
        lens_model_list = ["SIS"]
        redshift_list = [0.5]
        lensModelMutli = LensModel(
            z_source=z_source,
            lens_model_list=lens_model_list,
            lens_redshift_list=redshift_list,
            multi_plane=True,
        )
        lensModel = LensModel(lens_model_list=lens_model_list)
        kwargs_lens = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
        kappa_simple = lensModel.kappa(1, 0, kwargs_lens)
        kappa_multi = lensModelMutli.kappa(1, 0, kwargs_lens)
        npt.assert_almost_equal(kappa_simple, kappa_multi, decimal=5)

        gamma1_simple, gamma2_simple = lensModel.gamma(1, 0, kwargs_lens)
        gamma1_multi, gamma2_multi = lensModelMutli.gamma(1, 0, kwargs_lens)
        npt.assert_almost_equal(gamma1_simple, gamma1_multi, decimal=5)
        npt.assert_almost_equal(gamma2_simple, gamma2_multi, decimal=5)

        mag_simple = lensModel.magnification(0.99, 0, kwargs_lens)
        mag_multi = lensModelMutli.magnification(0.99, 0, kwargs_lens)
        npt.assert_almost_equal(mag_simple, mag_multi, decimal=5)

    def test_sis_travel_time(self):
        z_source = 1.5
        z_lens = 0.5
        lens_model_list = ["SIS"]
        redshift_list = [z_lens]
        lensModelMutli = MultiPlane(
            z_source=z_source,
            lens_model_list=lens_model_list,
            lens_redshift_list=redshift_list,
        )
        lensModel = LensModel(lens_model_list=lens_model_list)
        kwargs_lens = [{"theta_E": 1.0, "center_x": 0, "center_y": 0}]
        dt = lensModelMutli.arrival_time(1.0, 0.0, kwargs_lens)
        Dt = lensModelMutli._multi_plane_base._cosmo_bkg.ddt(
            z_lens=z_lens, z_source=z_source
        )
        fermat_pot = lensModel.fermat_potential(1, 0.0, kwargs_lens)
        dt_simple = const.delay_arcsec2days(fermat_pot, Dt)
        print(dt, dt_simple)
        npt.assert_almost_equal(dt, dt_simple, decimal=8)

    def test_sis_travel_time_new(self):
        z_source = 1.5
        z_lens = 0.5
        lens_model_list = ["SIS", "SIS"]
        redshift_list = [z_lens, 0.2]
        lensModelMutli = MultiPlane(
            z_source=z_source,
            lens_model_list=lens_model_list,
            lens_redshift_list=redshift_list,
        )
        lensModel = LensModel(lens_model_list=lens_model_list)
        kwargs_lens = [
            {"theta_E": 1.0, "center_x": 0, "center_y": 0},
            {"theta_E": 0.0, "center_x": 0, "center_y": 0},
        ]
        dt = lensModelMutli.arrival_time(1.0, 0.0, kwargs_lens)
        Dt = lensModelMutli._multi_plane_base._cosmo_bkg.ddt(
            z_lens=z_lens, z_source=z_source
        )
        fermat_pot = lensModel.fermat_potential(1, 0.0, kwargs_lens)
        dt_simple = const.delay_arcsec2days(fermat_pot, Dt)
        print(dt, dt_simple)
        npt.assert_almost_equal(dt, dt_simple, decimal=8)

    def test_sis_ray_shooting(self):
        z_source = 1.5
        z_lens = 0.5
        lens_model_list = ["SIS"]
        redshift_list = [z_lens]
        lensModelMutli = MultiPlane(
            z_source=z_source,
            lens_model_list=lens_model_list,
            lens_redshift_list=redshift_list,
        )
        lensModel = LensModel(lens_model_list=lens_model_list)
        kwargs_lens = [{"theta_E": 1.0, "center_x": 0, "center_y": 0}]
        beta_x, beta_y = lensModelMutli.ray_shooting(1.0, 0.0, kwargs_lens)
        beta_x_single, beta_y_single = lensModel.ray_shooting(1, 0.0, kwargs_lens)
        npt.assert_almost_equal(beta_x, beta_x_single, decimal=8)
        npt.assert_almost_equal(beta_y, beta_y_single, decimal=8)
        x, y = np.array([1.0]), np.array([2.0])
        beta_x, beta_y = lensModelMutli.ray_shooting(x, y, kwargs_lens)
        beta_x_single, beta_y_single = lensModel.ray_shooting(x, y, kwargs_lens)
        npt.assert_almost_equal(beta_x, beta_x_single, decimal=8)
        npt.assert_almost_equal(beta_y, beta_y_single, decimal=8)

    def test_random_ordering(self):
        z_source = 1.5
        lens_model_list = ["SIS", "SIS", "SIS"]
        sis1 = {"theta_E": 1.0, "center_x": 0, "center_y": 0}
        sis2 = {"theta_E": 0.2, "center_x": 0.5, "center_y": 0}
        sis3 = {"theta_E": 0.1, "center_x": 0, "center_y": 0.5}
        z1 = 0.1
        z2 = 0.5
        z3 = 0.7
        redshift_list = [z1, z2, z3]
        kwargs_lens = [sis1, sis2, sis3]
        lensModel = MultiPlane(
            z_source=z_source,
            lens_model_list=lens_model_list,
            lens_redshift_list=redshift_list,
        )
        beta_x_1, beta_y_1 = lensModel.ray_shooting(1.0, 0.0, kwargs_lens)

        redshift_list = [z3, z2, z1]
        kwargs_lens = [sis3, sis2, sis1]
        lensModel = MultiPlane(
            z_source=z_source,
            lens_model_list=lens_model_list,
            lens_redshift_list=redshift_list,
        )
        beta_x_2, beta_y_2 = lensModel.ray_shooting(1.0, 0.0, kwargs_lens)
        npt.assert_almost_equal(beta_x_1, beta_x_2, decimal=8)
        npt.assert_almost_equal(beta_y_1, beta_y_2, decimal=8)

    def test_ray_shooting_partial_2(self):
        z_source = 1.5
        lens_model_list = ["SIS", "SIS", "SIS", "SIS"]
        sis1 = {"theta_E": 0.4, "center_x": 0, "center_y": 0}
        sis2 = {"theta_E": 0.2, "center_x": 0.5, "center_y": 0}
        sis3 = {"theta_E": 0.1, "center_x": 0, "center_y": 0.5}
        sis4 = {"theta_E": 0.5, "center_x": 0.1, "center_y": 0.3}

        lens_model_list_macro = ["SIS"]
        kwargs_macro = [{"theta_E": 1, "center_x": 0, "center_y": 0}]

        zmacro = 0.5

        z1 = 0.1
        z2 = 0.5
        z3 = 0.5
        z4 = 0.7
        redshift_list = [z1, z2, z3, z4]
        kwargs_lens = [sis1, sis2, sis3, sis4]
        kwargs_lens_full = kwargs_macro + kwargs_lens
        lensModel_full = MultiPlane(
            z_source=z_source,
            lens_model_list=lens_model_list_macro + lens_model_list,
            lens_redshift_list=[zmacro] + redshift_list,
        )
        lensModel_macro = MultiPlane(
            z_source=z_source,
            lens_model_list=lens_model_list_macro,
            lens_redshift_list=[zmacro],
        )
        lensModel = MultiPlane(
            z_source=z_source,
            lens_model_list=lens_model_list,
            lens_redshift_list=redshift_list,
        )

        theta_x, theta_y = 1.0, 1.0

        x_subs, y_subs, alpha_x_subs, alpha_y_subs = lensModel.ray_shooting_partial(
            x=0,
            y=0,
            alpha_x=theta_x,
            alpha_y=theta_y,
            z_start=0,
            z_stop=zmacro,
            kwargs_lens=kwargs_lens,
        )

        x_out, y_out, alpha_x_out, alpha_y_out = lensModel_macro.ray_shooting_partial(
            x_subs,
            y_subs,
            alpha_x_subs,
            alpha_y_subs,
            zmacro,
            zmacro,
            kwargs_macro,
            include_z_start=True,
        )
        npt.assert_almost_equal(x_subs, x_out)
        npt.assert_almost_equal(y_subs, y_out)

        (
            x_full,
            y_full,
            alpha_x_full,
            alpha_y_full,
        ) = lensModel_full.ray_shooting_partial(
            0, 0, theta_x, theta_y, 0, zmacro, kwargs_lens_full
        )
        npt.assert_almost_equal(x_full, x_out)
        npt.assert_almost_equal(y_full, y_out)
        npt.assert_almost_equal(alpha_x_full, alpha_x_out)
        npt.assert_almost_equal(alpha_y_full, alpha_y_out)

        x_src, y_src, _, _ = lensModel_full.ray_shooting_partial(
            x=x_out,
            y=y_out,
            alpha_x=alpha_x_out,
            alpha_y=alpha_y_out,
            z_start=zmacro,
            z_stop=z_source,
            kwargs_lens=kwargs_lens_full,
        )

        beta_x, beta_y = lensModel.co_moving2angle_source(x_src, y_src)
        beta_x_true, beta_y_true = lensModel_full.ray_shooting(
            theta_x, theta_y, kwargs_lens_full
        )

        npt.assert_almost_equal(beta_x, beta_x_true, decimal=8)
        npt.assert_almost_equal(beta_y, beta_y_true, decimal=8)

    def test_ray_shooting_partial(self):
        z_source = 1.5
        lens_model_list = ["SIS", "SIS", "SIS"]
        sis1 = {"theta_E": 1.0, "center_x": 0, "center_y": 0}
        sis2 = {"theta_E": 0.2, "center_x": 0.5, "center_y": 0}
        sis3 = {"theta_E": 0.1, "center_x": 0, "center_y": 0.5}
        z1 = 0.1
        z2 = 0.5
        z3 = 0.7
        redshift_list = [z1, z2, z3]
        kwargs_lens = [sis1, sis2, sis3]
        lensModel = MultiPlane(
            z_source=z_source,
            lens_model_list=lens_model_list,
            lens_redshift_list=redshift_list,
        )
        lensModel_2 = LensModel(
            z_source=z_source,
            lens_model_list=lens_model_list,
            lens_redshift_list=redshift_list,
            multi_plane=True,
        )
        multiplane_2 = lensModel_2.lens_model
        intermediate_index = 1
        theta_x, theta_y = 1.0, 1.0

        Tzsrc = lensModel._multi_plane_base._cosmo_bkg.T_xy(0, z_source)

        z_intermediate = lensModel._multi_plane_base._lens_redshift_list[
            intermediate_index
        ]

        for lensmodel_class in [lensModel, multiplane_2]:
            (
                x_out,
                y_out,
                alpha_x_out,
                alpha_y_out,
            ) = lensmodel_class.ray_shooting_partial(
                x=0,
                y=0,
                alpha_x=theta_x,
                alpha_y=theta_y,
                z_start=0,
                z_stop=z_intermediate,
                kwargs_lens=kwargs_lens,
            )

            x_out_full_0 = x_out
            y_out_full_0 = y_out

            (
                x_out,
                y_out,
                alpha_x_out,
                alpha_y_out,
            ) = lensmodel_class.ray_shooting_partial(
                x=x_out,
                y=y_out,
                alpha_x=alpha_x_out,
                alpha_y=alpha_y_out,
                z_start=z_intermediate,
                z_stop=z_source,
                kwargs_lens=kwargs_lens,
            )

            beta_x, beta_y = lensModel.co_moving2angle_source(x_out, y_out)
            beta_x_true, beta_y_true = lensmodel_class.ray_shooting(
                theta_x, theta_y, kwargs_lens
            )
            npt.assert_almost_equal(beta_x, beta_x_true, decimal=8)
            npt.assert_almost_equal(beta_y, beta_y_true, decimal=8)

            T_ij_start = lensModel._multi_plane_base._cosmo_bkg.T_xy(
                z_observer=0, z_source=0.1
            )
            T_ij_end = lensModel._multi_plane_base._cosmo_bkg.T_xy(
                z_observer=0.7, z_source=1.5
            )
            (
                x_out,
                y_out,
                alpha_x_out,
                alpha_y_out,
            ) = lensmodel_class.ray_shooting_partial(
                x=0,
                y=0,
                alpha_x=theta_x,
                alpha_y=theta_y,
                z_start=0,
                z_stop=z_source,
                kwargs_lens=kwargs_lens,
                T_ij_start=T_ij_start,
                T_ij_end=T_ij_end,
            )

            beta_x, beta_y = x_out / Tzsrc, y_out / Tzsrc
            npt.assert_almost_equal(beta_x, beta_x_true, decimal=8)
            npt.assert_almost_equal(beta_y, beta_y_true, decimal=8)

    def test_pseudo_multiplane(self):
        z_source = 1.5
        lens_model_list = ["SIS", "SIS"]
        sis1 = {"theta_E": 1.0, "center_x": 0, "center_y": 0}
        sis2 = {"theta_E": 0.2, "center_x": 0.5, "center_y": 0}
        z1 = 0.5
        z2 = 0.5

        redshift_list = [z1, z2]
        kwargs_lens = [sis1, sis2]
        lensModelMulti = MultiPlane(
            z_source=z_source,
            lens_model_list=lens_model_list,
            lens_redshift_list=redshift_list,
        )
        lensModelSingle = LensModel(lens_model_list=lens_model_list)

        beta_x, beta_y = lensModelMulti.ray_shooting(1, 1, kwargs_lens)
        beta_x_single, beta_y_single = lensModelSingle.ray_shooting(1, 1, kwargs_lens)
        npt.assert_almost_equal(beta_x, beta_x_single, decimal=10)
        npt.assert_almost_equal(beta_y, beta_y_single, decimal=10)

    def test_position_convention(self):
        lens_model_list = ["SIS", "SIS", "SIS", "SIS"]
        redshift_list = [0.5, 0.5, 0.9, 0.6]

        kwargs_lens = [
            {"theta_E": 1, "center_x": 0, "center_y": 0},
            {"theta_E": 0.4, "center_x": 0, "center_y": 0.2},
            {"theta_E": 1, "center_x": 1.8, "center_y": -0.4},
            {"theta_E": 0.41, "center_x": 1.0, "center_y": 0.7},
        ]

        index_list = [[2, 3], [3, 2]]

        # compute the physical position given lensed position, and check that lensing computations
        # using the two different conventions and sets of kwargs agree

        for index in index_list:
            lensModel_observed = LensModel(
                lens_model_list=lens_model_list,
                multi_plane=True,
                observed_convention_index=index,
                z_source=1.5,
                lens_redshift_list=redshift_list,
            )
            lensModel_physical = LensModel(
                lens_model_list=lens_model_list,
                multi_plane=True,
                z_source=1.5,
                lens_redshift_list=redshift_list,
            )

            multi = lensModel_observed.lens_model._multi_plane_base
            lensed, phys = LensedLocation(multi, index), PhysicalLocation()

            kwargs_lens_physical = lensModel_observed.lens_model._convention(
                kwargs_lens
            )

            kwargs_phys, kwargs_lensed = phys(kwargs_lens), lensed(kwargs_lens)

            for j, lensed_kwargs in enumerate(kwargs_lensed):
                for ki in lensed_kwargs.keys():
                    assert lensed_kwargs[ki] == kwargs_lens_physical[j][ki]
                    assert kwargs_phys[j][ki] == kwargs_lens[j][ki]

            fxx, fyy, fxy, fyx = lensModel_observed.hessian(0.5, 0.5, kwargs_lens)
            fxx2, fyy2, fxy2, fyx2 = lensModel_physical.hessian(
                0.5, 0.5, kwargs_lens_physical
            )
            npt.assert_almost_equal(fxx, fxx2)
            npt.assert_almost_equal(fxy, fxy2)

            betax1, betay1 = lensModel_observed.ray_shooting(0.5, 0.5, kwargs_lens)
            betax2, betay2 = lensModel_physical.ray_shooting(
                0.5, 0.5, kwargs_lens_physical
            )
            npt.assert_almost_equal(betax1, betax2)
            npt.assert_almost_equal(betay1, betay2)

    def test_co_moving2angle_z1_z2(self):
        z_source = 1.5
        lens_model_list = ["SIS"]
        kwargs_lens = [{"theta_E": 1}]
        redshift_list = [0.5]
        lensModelMutli = MultiPlane(
            z_source=z_source,
            lens_model_list=lens_model_list,
            lens_redshift_list=redshift_list,
            z_interp_stop=3,
            cosmo_interp=False,
        )
        z1 = 0.5
        z2 = 1.5
        x, y = 1, 0
        beta_x, beta_y = lensModelMutli.co_moving2angle_z1_z2(x=x, y=y, z1=z1, z2=z2)
        T_ij_end = lensModelMutli._multi_plane_base._cosmo_bkg.T_xy(
            z_observer=z1, z_source=z2
        )
        npt.assert_almost_equal(beta_x, x / T_ij_end, decimal=7)
        npt.assert_almost_equal(beta_y, y / T_ij_end, decimal=7)

    def test_hessian_z1z2(self):
        z_source = 1.5
        lens_model_list = ["SIS"]
        kwargs_lens = [{"theta_E": 1}]
        redshift_list = [0.5]
        multi_plane = MultiPlane(
            z_source=z_source,
            lens_model_list=lens_model_list,
            lens_redshift_list=redshift_list,
            z_interp_stop=3,
            cosmo_interp=False,
        )
        z1, z2 = 0, z_source
        theta_x, theta_y = np.linspace(start=-1, stop=1, num=10), np.linspace(
            start=-1, stop=1, num=10
        )
        f_xx_z12, f_xy_z12, f_yx_z12, f_yy_z12 = multi_plane.hessian_z1z2(
            z1, z2, theta_x, theta_y, kwargs_lens, diff=0.00000001
        )
        f_xx, f_xy, f_yx, f_yy = multi_plane.hessian(
            theta_x, theta_y, kwargs_lens, diff=0.00000001
        )
        npt.assert_almost_equal(f_xx_z12, f_xx)
        npt.assert_almost_equal(f_xy_z12, f_xy)
        npt.assert_almost_equal(f_yx_z12, f_yx)
        npt.assert_almost_equal(f_yy_z12, f_yy)


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            MultiPlaneBase(
                z_source_convention=1, lens_model_list=["SIS"], lens_redshift_list=[2]
            )

        with self.assertRaises(ValueError):
            MultiPlaneBase(
                z_source_convention=1,
                lens_model_list=["SIS", "SIS"],
                lens_redshift_list=[0.5],
            )

        with self.assertRaises(ValueError):
            lens = MultiPlane(
                z_source_convention=1,
                z_source=1,
                lens_model_list=["SIS", "SIS"],
                lens_redshift_list=[0.5, 0.8],
            )
            lens._check_raise(k=[1])

        with self.assertRaises(ValueError):
            lens_model = LensModel(
                lens_model_list=["SIS"],
                multi_plane=True,
                z_source=1,
                z_source_convention=1,
                cosmo_interp=True,
                z_interp_stop=0.5,
            )


if __name__ == "__main__":
    pytest.main("-k TestLensModel")



import numpy as np
import lenstronomy.Util.param_util as param_util
import lenstronomy.Util.util as util
from scipy import signal
from lenstronomy.Util.kin_sampling_util import KinNNImageAlign
from lenstronomy.Sampling.Likelihoods import kinematic_NN_call

__all__ = ["KinLikelihood"]


class KinLikelihood(object):
    """Class to compute the likelihood associated to binned 2D kinematic maps."""

    def __init__(
        self,
        kinematic_bin_2d_class,
        lens_model_class,
        lens_light_model_class,
        kwargs_data,
        idx_lens=0,
        idx_lens_light=0,
    ):
        """
        :param kinematic_data_2d_class: KinBin class instance
        :param lens_model_class: LensModel class instance
        :param lens_light_model_class: LightModel class instance
        :param kwargs_data: kwargs describing image rotation
        :param idx_lens: int, index of the LensModel mass profile to consider for kinematics
        :param idx_lens_light: int, index of the lens LightModel profile to consider for kinematics
        """
        self.lens_model_class = lens_model_class
        self.z_lens = self.lens_model_class.z_lens
        self.lens_light_model_class = lens_light_model_class
        self._idx_lens = idx_lens
        self._idx_lens_light = idx_lens_light
        self.kin_class = kinematic_bin_2d_class

        # setup KinematicNN and KinNNImagealign class inputs to be updated when called
        self.kin_input = self.kin_class.kin_bin2kwargs()
        self.image_input = self.kwargs_data2image_input(kwargs_data)
        self.kinematic_NN = kinematic_NN_call.KinematicNN()
        self.config = self.kinematic_NN.config
        self.kinNN_input = {
            "deltaPix": self.config["grid_settings"]["deltapix"],
            "image": np.ones(
                (
                    self.config["grid_settings"]["npix"],
                    self.config["grid_settings"]["npix"],
                )
            ),
        }
        self.KiNNalign = KinNNImageAlign(
            self.kin_input, self.image_input, self.kinNN_input
        )

        self.data = self.kin_class.data
        self.psf = self.kin_class.PSF.kernel_point_source
        self.bin_mask = self.kin_class.bin_mask
        self.covariance = self.kin_class.covariance

        self.kin_x_grid, self.kin_y_grid = self.kin_class.kin_grid()

        self.lens_light_bool_list = list(
            np.zeros_like(self.lens_light_model_class.profile_type_list, dtype=bool)
        )
        self.lens_light_bool_list[self._idx_lens_light] = True

        d_d_fiducial = 1215.739  # fiducial distances used to create training set
        d_s_fiducial = 1650.753
        d_ds_fiducial = 1042.883
        z_d_fiducial = 0.5
        d_dt_fiducial = (1 + z_d_fiducial) * d_d_fiducial * d_s_fiducial / d_ds_fiducial
        self.fiducial_scale = d_dt_fiducial / (d_d_fiducial * (1 + z_d_fiducial))
        self.vrms = None

    def calc_vrms(self, kwargs_lens, kwargs_lens_light, kwargs_special, verbose=False):
        """Calculates binned vrms using SKiNN.

        :param kwargs_lens: lens model kwargs list
        :param kwargs_lens_light: lens light kwargs list
        :param kwargs_special: cosmology and other kwargs
        :param verbose: default False; if True print statements when out of bounds
            return binned vrms [km/s]; if SKiNN not installed return nan
        """
        self.update_image_input(kwargs_lens)
        self.light_map = self.lens_light_model_class.surface_brightness(
            self.kin_x_grid,
            self.kin_y_grid,
            kwargs_lens_light,
            self.lens_light_bool_list,
        )
        input_params, same_orientation = self.convert_to_nn_params(
            kwargs_lens, kwargs_lens_light, kwargs_special
        )
        if self.kinematic_NN.SKiNN_installed:
            velo_map = self.kinematic_NN.generate_map(input_params, verbose=verbose)
            velo_map = self.rescale_distance(
                velo_map, kwargs_special
            )  # RESCALE ACCORDING TO D_d, D_dt
            # Rotation and interpolation in kin data coordinates
            self.kinNN_input["image"] = velo_map
            self.KiNNalign.update(self.kin_input, self.image_input, self.kinNN_input)
            self.rotated_velo = self.KiNNalign.interp_image()
            # Convolution by PSF to calculate Vrms and binning
            vrms = self.auto_binning(self.rotated_velo, self.light_map)
            return vrms
        else:
            return np.nan

    def logL(self, kwargs_lens, kwargs_lens_light, kwargs_special, verbose=False):
        """Calculates Log likelihood from 2D kinematic likelihood.

        :param kwargs_lens: lens model kwargs list
        :param kwargs_lens_light: lens light kwargs list
        :param kwargs_special: cosmology and other kwargs
        :param verbose: default False; if True print statements when out of bounds
            return kinematics log likelihood
        """
        if self.kinematic_NN.SKiNN_installed:
            self.update_image_input(kwargs_lens)
            self.light_map = self.lens_light_model_class.surface_brightness(
                self.kin_x_grid,
                self.kin_y_grid,
                kwargs_lens_light,
                self.lens_light_bool_list,
            )
            input_params, same_orientation = self.convert_to_nn_params(
                kwargs_lens, kwargs_lens_light, kwargs_special
            )
            if not self.kinematic_NN.check_bounds(
                input_params, same_orientation=same_orientation, verbose=verbose
            ):
                # params not within training set. Penalty
                return -(10**8)
            self.vrms = self.calc_vrms(
                kwargs_lens, kwargs_lens_light, kwargs_special, verbose=verbose
            )
            logL = self._logL(self.vrms)
        else:
            logL = np.nan
        return logL

    def convert_to_nn_params(self, kwargs_lens, kwargs_lens_light, kwargs_special):
        """Converts lenstronomy kwargs into input vector for SKiNN, also returns whether
        or not mass and light are aligned.

        :param kwargs_lens: lens model kwargs list
        :param kwargs_lens_light: lens light kwargs list
        :param kwargs_special: cosmology and other kwargs
        :return: parameters in GLEE convention to be input into NN
        """
        # lenstronomy to GLEE conversion
        # orientation_mass,q_mass=param_util.ellipticity2phi_q(kwargs_lens[self._idx_lens]['e1'],
        #                                                      kwargs_lens[self._idx_lens]['e2'])
        orientation_mass = kwargs_lens[self._idx_lens]["phi"]
        q_mass = kwargs_lens[self._idx_lens]["q"]
        # orientation_light, q_light = param_util.ellipticity2phi_q(kwargs_lens_light[self._idx_lens]['e1'],
        #                                                         kwargs_lens_light[self._idx_lens]['e2'])
        orientation_light = kwargs_lens_light[self._idx_lens]["phi"]
        q_light = kwargs_lens_light[self._idx_lens]["q"]

        same_orientation = True  # confirm that angles are aligned
        if np.abs(orientation_mass - orientation_light) > 0.05:
            if q_light < 0.95 or q_mass < 0.95:
                same_orientation = False

        theta_e_lenstro = kwargs_lens[self._idx_lens]["theta_E"]
        gamma_lenstro = kwargs_lens[self._idx_lens]["gamma"]
        gamma_glee = (gamma_lenstro - 1) / 2
        r_e_scale = (2 / (1 + q_mass)) ** (1 / (2 * gamma_glee)) * np.sqrt(q_mass)
        theta_e_glee = theta_e_lenstro / r_e_scale
        return (
            np.array(
                [
                    q_mass,
                    q_light,
                    theta_e_glee,
                    kwargs_lens_light[self._idx_lens]["n_sersic"],
                    kwargs_lens_light[self._idx_lens]["R_sersic"],
                    8.0e-2,
                    gamma_glee,
                    kwargs_special["b_ani"],
                    kwargs_special["incli"] * 180 / np.pi,
                ]
            ),
            same_orientation,
        )

    def rescale_distance(self, image, kwargs_special):
        """Rescales velocity map according to distance, requires lens redshift.

        :param image: vrms image [km/s]
        :param kwargs_special: kwargs with cosmological distances for rescaling return
            rescaled vrms image [km/s]
        """
        new_scale = kwargs_special["D_dt"] / (kwargs_special["D_d"] * (1 + self.z_lens))
        factor = np.sqrt(
            new_scale / self.fiducial_scale
        )  # vrms^2 propto D_dt/Dd (see https://arxiv.org/pdf/2109.14615.pdf Eq. 20)
        return image * factor

    def kwargs_data2image_input(self, kwargs_data):
        """Creates the kwargs of the image needed for 2D kinematic likelihood.

        :param kwargs_data: kwargs giving image and describing imaging data coordinate
            transformation
        :return kwargs: coordinate transformation kwargs as input for KinNNImageAlign
            class
        """
        delta_pix = np.sqrt(np.abs(np.linalg.det(kwargs_data["transform_pix2angle"])))
        kwargs = {
            "image": kwargs_data["image_data"],
            "deltaPix": delta_pix,
            "transform_pix2angle": kwargs_data["transform_pix2angle"],
            "ra_at_xy0": kwargs_data["ra_at_xy_0"],
            "dec_at_xy0": kwargs_data["dec_at_xy_0"],
        }
        return kwargs

    def update_image_input(self, kwargs_lens):
        """Updates the image_input for rotation with the new values of orientation and
        center of the lens model.

        :param kwargs_lens: lens kwargs with center and PA positions to be used as input
        """
        orientation_ellipse = kwargs_lens[self._idx_lens]["phi"]
        q = kwargs_lens[self._idx_lens]["q"]
        cx = kwargs_lens[self._idx_lens]["center_x"]
        cy = kwargs_lens[self._idx_lens]["center_y"]
        self.image_input["ellipse_PA"] = orientation_ellipse
        self.image_input["offset_x"] = cx
        self.image_input["offset_y"] = cy

    def auto_binning(self, rotated_map, light_map):
        """Function to convolve and bin the NN rotated output.

        :param rotated_map: model vrms map [km/s] in data pixel coordinates
        :param light_map: model light map in data pixel coordinates for weighting
        :return: binned vrms [km/s] for comparison with data
        """
        vrms = rotated_map
        mge_car = light_map

        vrms = signal.fftconvolve(vrms, self.psf, mode="same")
        mge_car_con = signal.fftconvolve(mge_car, self.psf, mode="same")

        numerator = []
        denominator = []

        for idx in range(len(self.data)):
            math_pos = np.where(self.bin_mask == idx)
            if np.shape(math_pos[0])[0] == 0:
                raise ValueError(
                    "binmap mismatch with data: no pixels in bin with idx %i" % (idx)
                )
            numerator.append(np.sum(vrms[math_pos] * mge_car_con[math_pos]))
            denominator.append(np.sum(mge_car_con[math_pos]))

        vrms = np.array(numerator) / np.array(denominator).clip(0)

        return vrms

    def _logL(self, vrms):
        """Calculates the log likelihood for a given binned model.

        :param vrms: binned vrms [km/s] to compare with observed binned data
        :return: log likelihood
        """
        # log_like = (vrms - self.data)**2 / self.noise**2
        cov_inv = np.linalg.inv(self.covariance)
        log_like = np.matmul(
            np.matmul((vrms - self.data).T, cov_inv), (vrms - self.data)
        )
        logL = -np.sum(log_like) / 2

        if not np.isfinite(logL):
            return -(10**15)

        return logL



__author__ = "sibirrer"

import numpy as np
import math
import scipy.special

import lenstronomy.Util.param_util as param_util

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
class ShapeletsPolar(object):
    """2D polar Shapelets, see Massey & Refregier 2005."""

    param_names = ["amp", "beta", "n", "m", "center_x", "center_y"]
    param_names_latex = {r"$I_0$", r"$\beta$", r"$n$", r"$m$", r"$x_0$", r"$y_0$"}
    lower_limit_default = {
        "amp": 0,
        "beta": 0,
        "n": 0,
        "m": 0,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 100,
        "beta": 100,
        "n": 150,
        "m": 150,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        """Load interpolation of the Hermite polynomials in a range [-30,30] in order
        n<= 150 :return:"""
        pass

    def function(self, x, y, amp, beta, n, m, complex_bool, center_x, center_y):
        """

        :param x: x-coordinate, numpy array
        :param y: y-ccordinate, numpy array
        :param amp: amplitude normalization
        :param beta: shaplet scale
        :param n: order of polynomial
        :param m: rotational invariance
        :param complex_bool: boolean; if True uses complex value of function _chi_n_m()
        :param center_x: center of shapelet
        :param center_y: center of shapelet
        :return: amplitude of shapelet at possition (x, y)
        """
        r, phi = param_util.cart2polar(x, y, center_x, center_y)
        if complex_bool is True:
            return amp * self._chi_n_m(r, beta, n, m) * np.exp(-1j * m * phi).imag
        else:
            return amp * self._chi_n_m(r, beta, n, m) * np.exp(-1j * m * phi).real

    @staticmethod
    def _chi_n_m(r, beta, n, m):
        """

        :param r: radius / beta
        :param beta: shapelet scale
        :param n: non-negative integer
        :param m: integer, running from -n to n in steps of two
        :return: value of function (8) in Massey & Refregier, complex numbers
        """
        m_abs = int(abs(m))
        p = int((n - m_abs) / 2)
        p2 = int((n + m_abs) / 2)
        if p % 2 == 0:  # if p is even
            prefac = 1
        else:
            prefac = -1
        prefactor = (
            prefac
            / beta ** (m_abs + 1)
            * np.sqrt(math.factorial(p) / (np.pi * math.factorial(p2)))
        )
        poly = scipy.special.genlaguerre(
            n=p, alpha=m_abs
        )  # lower part, upper part of definition in Massey & Refregier
        r_ = (r / beta) ** 2
        l_n_alpha = poly(r_)
        return prefactor * r**m_abs * l_n_alpha * np.exp(-((r / beta) ** 2) / 2)

    @staticmethod
    def _index2n(index):
        """

        :param index: index of convention
        :return: n order of Laguerre
        """
        n_float = (-3 + np.sqrt(9 + 8 * index)) / 2
        n_int = int(n_float)
        if n_int == n_float:
            n = n_int
        else:
            n = n_int + 1
        return n

    def index2poly(self, index):
        """Manages the convention from an iterative index to the specific polynomial n,
        m, (real/imaginary part)

        :param index: int, index of list
        :return: n, m bool
        """

        n = self._index2n(index)

        num_prev = n * (n + 1) / 2
        num = index + 1
        delta = int(num - num_prev - 1)
        if n % 2 == 0:
            if delta == 0:
                m = delta
                complex_bool = False
            elif delta % 2 == 0:
                complex_bool = True
                m = delta
            else:
                complex_bool = False
                m = delta + 1
        else:
            if delta % 2 == 0:
                complex_bool = False
                m = delta + 1
            else:
                complex_bool = True
                m = delta
        return n, m, complex_bool

    @staticmethod
    def poly2index(n, m, complex_bool):
        """

        :param n: non-negative integer
        :param m: integer, running from -n to n in steps of two
        :param complex_bool: bool, if True, assigns complex part
        :return:
        """
        index = n * (n + 1) / 2
        if complex_bool is True:
            if m == 0:
                raise ValueError("m=0 can not have imaginary part!")
        if n % 2 == 0:
            if m % 2 == 0:
                if m == 0:
                    index += 1
                else:
                    index += m
                if complex_bool is True:
                    index += 1
            else:
                raise ValueError("m needs to be even for even n!")
        else:
            if complex_bool is True:
                index += m + 1
            else:
                index += m
        return int(index - 1)

    @staticmethod
    def num_param(n_max):
        """

        :param n_max: maximal polynomial order
        :return: number of basis components
        """
        return int((n_max + 1) * (n_max + 2) / 2)


@export
class ShapeletsPolarExp(object):
    """2D exponential shapelets, Berge et al.

    2019
    """

    param_names = ["amp", "beta", "n", "m", "center_x", "center_y"]
    lower_limit_default = {
        "amp": 0,
        "beta": 0,
        "n": 0,
        "m": 0,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 100,
        "beta": 100,
        "n": 150,
        "m": 150,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        """Load interpolation of the Hermite polynomials in a range [-30,30] in order
        n<= 150 :return:"""
        pass

    def function(self, x, y, amp, beta, n, m, complex_bool, center_x, center_y):
        """

        :param x: x-coordinate, numpy array
        :param y: y-ccordinate, numpy array
        :param amp: amplitude normalization
        :param beta: shaplet scale
        :param n: order of polynomial
        :param m: rotational invariance
        :param complex_bool: boolean; if True uses complex value of function _chi_n_m()
        :param center_x: center of shapelet
        :param center_y: center of shapelet
        :return: amplitude of shapelet at possition (x, y)
        """
        r, phi = param_util.cart2polar(x, y, center_x, center_y)
        if complex_bool is True:
            return amp * self._chi_n_m(r, beta, n, m) * np.exp(-1j * m * phi).imag
        else:
            return amp * self._chi_n_m(r, beta, n, m) * np.exp(-1j * m * phi).real

    @staticmethod
    def _chi_n_m(r, beta, n, m):
        """

        :param r: radius
        :param beta: shapelet scale
        :param n: non-negative integer
        :param m: integer, running from -n to n in steps of two
        :return: value of function (20) in Berge et al. 2019 without complex coefficient
        """
        m_abs = int(abs(m))
        p = int(n - m_abs)
        p2 = int(n + m_abs)
        if p % 2 == 0:  # if p is even
            prefac = 1
        else:
            prefac = -1
        prefactor = prefac * np.sqrt(
            2.0
            / (beta * np.pi * (2 * n + 1) ** 3)
            * math.factorial(p)
            / math.factorial(p2)
        )
        poly = scipy.special.genlaguerre(
            n=p, alpha=2 * m_abs
        )  # lower part, upper part of definition in Massey & Refregier
        x_ = 2.0 * r / (beta * (2 * n + 1))
        l_n_alpha = poly(x_)
        return prefactor * x_**m_abs * l_n_alpha * np.exp(-x_ / 2) / np.sqrt(beta)

    @staticmethod
    def num_param(n_max):
        """

        :param n_max: maximal polynomial order
        :return: number of basis components
        """
        return int((n_max + 1) ** 2)

    @staticmethod
    def _index2n(index):
        """

        :param index: index convention
        :return: n
        """
        n_float = np.sqrt(index + 1) - 1
        n_int = int(n_float)
        if n_int == n_float:
            n = n_int
        else:
            n = n_int + 1
        return n

    def index2poly(self, index):
        """

        :param index:
        :return:
        """
        n = self._index2n(index)
        num_prev = n**2
        delta = index + 1 - num_prev
        if delta % 2 == 1:
            m = int((delta - 1) / 2)
            complex_bool = False
        else:
            m = int(delta / 2)
            complex_bool = True
        return n, m, complex_bool

    @staticmethod
    def poly2index(n, m, complex_bool):
        """

        :param n:
        :param m:
        :param complex_bool:
        :return: index convention, integer
        """
        index = n**2
        index += 2 * m
        if complex_bool is True:
            index -= 1
        return int(index)


@export
class ShapeletSetPolar(object):
    """Class to operate on entire shapelet set."""

    param_names = ["amp", "n_max", "beta", "center_x", "center_y"]
    lower_limit_default = {"beta": 0, "center_x": -100, "center_y": -100}
    upper_limit_default = {"beta": 100, "center_x": 100, "center_y": 100}

    def __init__(self, exponential=False):
        if exponential is True:
            self.shapelets = ShapeletsPolarExp()
        else:
            self.shapelets = ShapeletsPolar()
        self._exponential = exponential

    def function(self, x, y, amp, n_max, beta, center_x=0, center_y=0):
        """

        :param x:
        :param y:
        :param amp:
        :param n_max:
        :param beta:
        :param center_x:
        :param center_y:
        :return:
        """
        num_param = self.shapelets.num_param(n_max)
        f_ = np.zeros(len(np.atleast_1d(x)))
        l_list = self._pre_calc(x, y, beta, n_max, center_x, center_y)
        for i in range(num_param):
            out = self._pre_calc_function(l_list, i) * amp[i]
            f_ += out
        try:
            len(x)
        except:
            f_ = f_[0]
        return np.nan_to_num(f_)

    def function_split(self, x, y, amp, n_max, beta, center_x=0, center_y=0):
        num_param = self.shapelets.num_param(n_max)
        A = []
        L_list = self._pre_calc(x, y, beta, n_max, center_x, center_y)
        for i in range(num_param):
            A.append(self._pre_calc_function(L_list, i) * amp[i])
        return A

    def _pre_calc(self, x, y, beta, n_max, center_x, center_y):
        """

        :param x:
        :param y:
        :param beta:
        :param n_max:
        :param center_x:
        :param center_y:
        :return:
        """
        L_list = []
        # polar coordinates
        r, phi = param_util.cart2polar(x, y, center_x, center_y)
        num_param = self.shapelets.num_param(n_max)

        # compute real and imaginary part of the complex angles in the range n_max
        theta_m_real_list, theta_m_imag_list = [], []
        for i in range(n_max + 1):
            m = i
            exp_complex = np.exp(-1j * m * phi)
            theta_m_real_list.append(exp_complex.real)
            theta_m_imag_list.append(exp_complex.imag)

        # compute the Laguerre polynomials in n, m
        chi_n_m_list = [[0 for _ in range(n_max + 1)] for _ in range(n_max + 1)]
        for n in range(n_max + 1):
            for m in range(n + 1):
                if (n - m) % 2 == 0 or self._exponential is True:
                    chi_n_m_list[n][m] = self.shapelets._chi_n_m(r, beta, n, m)

        # combine together the pre-computed components
        for index in range(num_param):
            n, m, complex_bool = self.shapelets.index2poly(index)
            if complex_bool is True:
                L_i = chi_n_m_list[n][m] * theta_m_imag_list[m]
            else:
                L_i = chi_n_m_list[n][m] * theta_m_real_list[m]
            L_list.append(L_i)
        return L_list

    @staticmethod
    def _pre_calc_function(L_list, i):
        """Evaluates the shapelet function based on the pre-calculated components in
        _pre_calc()

        :param L_list: pre-calculated components
        :param i: index conventions of the sequence of basis components
        :return: shaplet basis at the pre-calculated positions
        """
        return L_list[i]

    def decomposition(self, image, x, y, n_max, beta, deltaPix, center_x=0, center_y=0):
        """Decomposes an image into the shapelet coefficients in same order as for the
        function call :param image:

        :param x:
        :param y:
        :param n_max:
        :param beta:
        :param center_x:
        :param center_y:
        :return:
        """
        num_param = self.shapelets.num_param(n_max)
        param_list = np.zeros(num_param)
        amp_norm = 1.0 * deltaPix**2
        L_list = self._pre_calc(x, y, beta, n_max, center_x, center_y)
        for i in range(num_param):
            base = self._pre_calc_function(L_list, i) * amp_norm
            param = np.sum(image * base)
            n, m, complex_bool = self.shapelets.index2poly(i)
            if m != 0:
                param *= 2
            param_list[i] = param
        return param_list

    def index2poly(self, index):
        """

        :param index: index of coefficient in the convention here
        :return: n, m, complex_bool
        """
        return self.shapelets.index2poly(index)



__author__ = "sibirrer"

import numpy as np
from lenstronomy.LensModel.profile_list_base import ProfileListBase

__all__ = ["SinglePlane"]


class SinglePlane(ProfileListBase):
    """Class to handle an arbitrary list of lens models in a single lensing plane."""

    def ray_shooting(self, x, y, kwargs, k=None):
        """Maps image to source position (inverse deflection).

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the
            lens model classes
        :param k: only evaluate the k-th lens model
        :return: source plane positions corresponding to (x, y) in the image plane
        """

        dx, dy = self.alpha(x, y, kwargs, k=k)
        return x - dx, y - dy

    def fermat_potential(
        self, x_image, y_image, kwargs_lens, x_source=None, y_source=None, k=None
    ):
        """Fermat potential (negative sign means earlier arrival time)

        :param x_image: image position
        :param y_image: image position
        :param x_source: source position
        :param y_source: source position
        :param kwargs_lens: list of keyword arguments of lens model parameters matching
            the lens model classes
        :param k:
        :return: fermat potential in arcsec**2 without geometry term (second part of Eqn
            1 in Suyu et al. 2013) as a list
        """

        potential = self.potential(x_image, y_image, kwargs_lens, k=k)
        if x_source is None or y_source is None:
            x_source, y_source = self.ray_shooting(x_image, y_image, kwargs_lens, k=k)
        geometry = ((x_image - x_source) ** 2 + (y_image - y_source) ** 2) / 2.0
        return geometry - potential

    def potential(self, x, y, kwargs, k=None):
        """Lensing potential.

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the
            lens model classes
        :param k: only evaluate the k-th lens model
        :return: lensing potential in units of arcsec^2
        """
        x = np.array(x, dtype=float)
        y = np.array(y, dtype=float)
        if isinstance(k, int):
            return self.func_list[k].function(x, y, **kwargs[k])
        bool_list = self._bool_list(k)
        potential = np.zeros_like(x)
        for i, func in enumerate(self.func_list):
            if bool_list[i] is True:
                potential += func.function(x, y, **kwargs[i])
        return potential

    def alpha(self, x, y, kwargs, k=None):
        """Deflection angles.

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the
            lens model classes
        :param k: only evaluate the k-th lens model
        :return: deflectionangles in units of arcsec
        """
        x = np.array(x, dtype=float)
        y = np.array(y, dtype=float)

        if isinstance(k, int):
            return self.func_list[k].derivatives(x, y, **kwargs[k])
        bool_list = self._bool_list(k)
        f_x, f_y = np.zeros_like(x), np.zeros_like(x)
        for i, func in enumerate(self.func_list):
            if bool_list[i] is True:
                f_x_i, f_y_i = func.derivatives(x, y, **kwargs[i])
                f_x += f_x_i
                f_y += f_y_i

        return f_x, f_y

    def hessian(self, x, y, kwargs, k=None):
        """Hessian matrix.

        :param x: x-position (preferentially arcsec)
        :type x: numpy array
        :param y: y-position (preferentially arcsec)
        :type y: numpy array
        :param kwargs: list of keyword arguments of lens model parameters matching the
            lens model classes
        :param k: only evaluate the k-th lens model
        :return: f_xx, f_xy, f_yx, f_yy components
        """
        x = np.array(x, dtype=float)
        y = np.array(y, dtype=float)
        if isinstance(k, int):
            f_xx, f_xy, f_yx, f_yy = self.func_list[k].hessian(x, y, **kwargs[k])
            return f_xx, f_xy, f_yx, f_yy

        bool_list = self._bool_list(k)
        f_xx, f_xy, f_yx, f_yy = (
            np.zeros_like(x),
            np.zeros_like(x),
            np.zeros_like(x),
            np.zeros_like(x),
        )
        for i, func in enumerate(self.func_list):
            if bool_list[i] is True:
                f_xx_i, f_xy_i, f_yx_i, f_yy_i = func.hessian(x, y, **kwargs[i])
                f_xx += f_xx_i
                f_xy += f_xy_i
                f_yx += f_yx_i
                f_yy += f_yy_i
        return f_xx, f_xy, f_yx, f_yy

    def mass_3d(self, r, kwargs, bool_list=None):
        """Computes the mass within a 3d sphere of radius r.

        if you want to have physical units of kg, you need to multiply by this factor:
        const.arcsec ** 2 * self._cosmo.dd * self._cosmo.ds / self._cosmo.dds *
        const.Mpc * const.c ** 2 / (4 * np.pi * const.G) grav_pot = -const.G * mass_dim
        / (r * const.arcsec * self._cosmo.dd * const.Mpc)

        :param r: radius (in angular units)
        :param kwargs: list of keyword arguments of lens model parameters matching the
            lens model classes
        :param bool_list: list of bools that are part of the output
        :return: mass (in angular units, modulo epsilon_crit)
        """
        bool_list = self._bool_list(bool_list)
        mass_3d = 0
        for i, func in enumerate(self.func_list):
            if bool_list[i] is True:
                kwargs_i = {
                    k: v
                    for k, v in kwargs[i].items()
                    if k not in ["center_x", "center_y"]
                }
                mass_3d_i = func.mass_3d_lens(r, **kwargs_i)
                mass_3d += mass_3d_i
        return mass_3d

    def mass_2d(self, r, kwargs, bool_list=None):
        """Computes the mass enclosed a projected (2d) radius r.

        The mass definition is such that:

        .. math::
            \\alpha = mass_2d / r / \\pi

        with alpha is the deflection angle

        :param r: radius (in angular units)
        :param kwargs: list of keyword arguments of lens model parameters matching the
            lens model classes
        :param bool_list: list of bools that are part of the output
        :return: projected mass (in angular units, modulo epsilon_crit)
        """
        bool_list = self._bool_list(bool_list)
        mass_2d = 0
        for i, func in enumerate(self.func_list):
            if bool_list[i] is True:
                kwargs_i = {
                    k: v
                    for k, v in kwargs[i].items()
                    if k not in ["center_x", "center_y"]
                }
                mass_2d_i = func.mass_2d_lens(r, **kwargs_i)
                mass_2d += mass_2d_i
        return mass_2d

    def density(self, r, kwargs, bool_list=None):
        """3d mass density at radius r The integral in the LOS projection of this
        quantity results in the convergence quantity.

        :param r: radius (in angular units)
        :param kwargs: list of keyword arguments of lens model parameters matching the
            lens model classes
        :param bool_list: list of bools that are part of the output
        :return: mass density at radius r (in angular units, modulo epsilon_crit)
        """
        bool_list = self._bool_list(bool_list)
        density = 0
        for i, func in enumerate(self.func_list):
            if bool_list[i] is True:
                kwargs_i = {
                    k: v
                    for k, v in kwargs[i].items()
                    if k not in ["center_x", "center_y"]
                }
                density_i = func.density_lens(r, **kwargs_i)
                density += density_i
        return density



import lenstronomy.Util.param_util as param_util

__all__ = ["PJaffe", "PJaffeEllipse"]


class PJaffe(object):
    """Class for pseudo Jaffe lens light (2d projected light/mass distribution)"""

    param_names = ["amp", "Ra", "Rs", "center_x", "center_y"]
    lower_limit_default = {
        "amp": 0,
        "Ra": 0,
        "Rs": 0,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 100,
        "Ra": 100,
        "Rs": 100,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        from lenstronomy.LensModel.Profiles.p_jaffe import PJaffe as PJaffe_lens

        self.lens = PJaffe_lens()

    def function(self, x, y, amp, Ra, Rs, center_x=0, center_y=0):
        """

        :param x:
        :param y:
        :param amp:
        :param Ra:
        :param Rs:
        :param center_x:
        :param center_y:
        :return:
        """
        rho0 = self.lens.sigma2rho(amp, Ra, Rs)
        return self.lens.density_2d(x, y, rho0, Ra, Rs, center_x, center_y)

    def light_3d(self, r, amp, Ra, Rs):
        """

        :param r:
        :param amp:
        :param Rs:
        :param Ra:
        :return:
        """
        rho0 = self.lens.sigma2rho(amp, Ra, Rs)
        return self.lens.density(r, rho0, Ra, Rs)


class PJaffeEllipse(object):
    """Calss for elliptical pseudo Jaffe lens light."""

    param_names = ["amp", "Ra", "Rs", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "amp": 0,
        "Ra": 0,
        "Rs": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "amp": 100,
        "Ra": 100,
        "Rs": 100,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        from lenstronomy.LensModel.Profiles.p_jaffe import PJaffe as PJaffe_lens

        self.lens = PJaffe_lens()
        self.spherical = PJaffe()

    def function(self, x, y, amp, Ra, Rs, e1, e2, center_x=0, center_y=0):
        """

        :param x:
        :param y:
        :param amp:
        :param Ra:
        :param Rs:
        :param center_x:
        :param center_y:
        :return:
        """
        x_, y_ = param_util.transform_e1e2_square_average(
            x, y, e1, e2, center_x, center_y
        )
        return self.spherical.function(x_, y_, amp, Ra, Rs)

    def light_3d(self, r, amp, Ra, Rs, e1=0, e2=0):
        """

        :param r:
        :param amp:
        :param Ra:
        :param Rs:
        :return:
        """
        return self.spherical.light_3d(r, amp, Ra, Rs)
